# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
from pprint import pformat
from typing import Any, ClassVar, Dict, List, Optional, Union

import attr

from .label import Label
from .performance import Performance
from .task import Task
from .utils import (
    attr_value_serializer,
    deidentify,
    remove_null_fields,
    str_to_datetime,
)


@attr.define
class Connection:
    """
    Representation of a connection between two tasks in GETi.

    :var to: Name of the task to which the connection is leading
    :var from_: Name of the task from which the connection originates
    """

    to: str
    from_: str


@attr.define
class Pipeline:
    """
    Representation of a pipeline for a project in GETi.

    :var tasks: List of tasks in the pipeline
    :var connections: List of connections between the tasks in the pipeline
    """

    tasks: List[Task]
    connections: List[Connection]

    @property
    def trainable_tasks(self) -> List[Task]:
        """
        Return an ordered list of trainable tasks.

        :return: List of trainable tasks in the pipeline
        """
        return [task for task in self.tasks if task.is_trainable]

    def resolve_connections(self):
        """
        Replace the task id's in the connections attribute of the pipeline with the
        actual task titles.

        :return:
        """
        task_id_to_name = {task.id: task.title for task in self.tasks}
        for connection in self.connections:
            if connection.to in task_id_to_name.keys():
                connection.to = task_id_to_name[connection.to]
            if connection.from_ in task_id_to_name.keys():
                connection.from_ = task_id_to_name[connection.from_]

    def resolve_parent_labels(self):
        """
        Replace the parent_id's in the pipeline labels by the actual label names.

        :return:
        """
        for task in self.trainable_tasks:
            for label in task.labels:
                if label.parent_id in self.label_id_to_name_mapping.keys():
                    label.parent_id = self.label_id_to_name_mapping[label.parent_id]

    @property
    def label_id_to_name_mapping(self) -> Dict[str, str]:
        """
        Return a mapping of label ID's to label names for all labels in the pipeline.

        :return: dictionary containing the label ID's as keys and the label names as
            values
        """
        label_mapping: Dict[str, str] = {}
        for task in self.trainable_tasks:
            label_mapping.update({label.id: label.name for label in task.labels})
        return label_mapping

    def get_labels_per_task(self, include_empty: bool = True) -> List[List[Label]]:
        """
        Return a nested list of labels for each task in the pipeline.

        Each entry in the outermost list corresponds to a task in the pipeline. The
        innermost list holds the labels for that specific task

        :param include_empty: True to include empty labels in the output
        :return: nested list of labels for each task in the pipeline
        """
        outer_list: List[List[Label]] = []
        for task in self.trainable_tasks:
            if not include_empty:
                task_labels = [label for label in task.labels if not label.is_empty]
            else:
                task_labels = task.labels
            outer_list.append(task_labels)
        return outer_list

    def get_all_labels(self) -> List[Label]:
        """
        Return a list of all labels in the pipeline.

        :return: List of all labels for every task in the pipeline
        """
        label_list: List[Label] = []
        for task in self.trainable_tasks:
            label_list.extend(task.labels)
        return label_list

    def deidentify(self) -> None:
        """
        Remove all unique database ID's from the tasks and labels in the pipeline.
        """
        self.resolve_connections()
        self.resolve_parent_labels()
        for task in self.tasks:
            task.deidentify()

    def prepare_for_post(self) -> None:
        """
        Set all fields to None that are not valid for making a POST request to the
        /projects endpoint.

        :return:
        """
        for task in self.tasks:
            task.prepare_for_post()


@attr.define
class Dataset:
    """
    Representation of a dataset for a project in Intel® Geti™.

    :var id: Unique database ID of the dataset
    :var name: name of the dataset
    """

    _identifier_fields: ClassVar[str] = ["id", "creation_time"]
    _GET_only_fields: ClassVar[List[str]] = ["use_for_training", "creation_time"]

    name: str
    id: Optional[str] = None
    creation_time: Optional[str] = attr.field(default=None, converter=str_to_datetime)
    use_for_training: Optional[bool] = None

    def deidentify(self) -> None:
        """
        Remove unique database ID from the Dataset.
        """
        deidentify(self)

    def prepare_for_post(self) -> None:
        """
        Set all fields to None that are not valid for making a POST request to the
        /projects endpoint.

        :return:
        """
        for field_name in self._GET_only_fields:
            setattr(self, field_name, None)


@attr.define
class Project:
    """
    Representation of a project in Intel® Geti™.

    :var id: Unique database ID of the project
    :var name: Name of the project
    :var creation_time: Time at which the project was created
    :var pipeline: Pipeline for the project
    :var datasets: List of datasets belonging to the project
    :var score: Score achieved by the AI assistant for the project
    :var thumbnail: URL at which a thumbnail for the project can be obtained
    :var storage_info: Storage information for the project can be obtained
    """

    _identifier_fields: ClassVar[List[str]] = [
        "id",
        "thumbnail",
        "creation_time",
        "creator_id",
    ]
    _GET_only_fields: ClassVar[List[str]] = [
        "thumbnail",
        "score",
        "performance",
        "creator_id",
        "creation_time",
        "storage_info",
    ]

    name: str
    pipeline: Pipeline
    datasets: Optional[List[Dataset]] = (
        None  # `datasets` was removed from project listing in Geti v1.15
    )
    score: Optional[float] = None  # 'score' is removed in v1.1
    performance: Optional[Performance] = None
    creation_time: Optional[str] = attr.field(default=None, converter=str_to_datetime)
    id: Optional[str] = None
    thumbnail: Optional[str] = None
    creator_id: Optional[str] = None
    storage_info: Optional[Dict] = None

    def get_trainable_tasks(self) -> List[Task]:
        """
        Return an ordered list of trainable tasks.

        :return: List of trainable tasks in the project
        """
        return self.pipeline.trainable_tasks

    @property
    def project_type(self) -> str:
        """
        Return a string containing the project type. The type is constructed by
        concatenating the types of tasks in the task chain together, separated by a
        '_to_' sequence.

        For example, a project with a detection task followed by a segmentation task
        will have it's project type set as: 'detection_to_segmentation'

        :return: string containing the project type
        """
        tasks = self.get_trainable_tasks()
        project_type = ""
        for index, task in enumerate(tasks):
            if index > 0:
                project_type += f"_to_{task.task_type}"
            else:
                project_type += f"{task.task_type}"
        return project_type

    def deidentify(self) -> None:
        """
        Remove all unique database ID's from the project, pipeline and datasets.

        :return:
        """
        deidentify(self)
        self.pipeline.deidentify()
        for dataset in self.datasets:
            dataset.deidentify()

    def prepare_for_post(self) -> None:
        """
        Set all fields to None that are not valid for making a POST request to the
        /projects endpoint.

        :return:
        """
        for field_name in self._GET_only_fields:
            setattr(self, field_name, None)
        self.pipeline.prepare_for_post()
        for dataset in self.datasets:
            dataset.prepare_for_post()

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the project to a dictionary representation.

        :return: Dictionary holding the project data
        """
        output_dict = attr.asdict(self, value_serializer=attr_value_serializer)
        for connection in output_dict["pipeline"]["connections"]:
            from_value = connection.pop("from_")
            connection["from"] = from_value
        return output_dict

    def get_labels_per_task(
        self, include_empty: bool = True
    ) -> List[List[Dict[str, Any]]]:
        """
        Return a nested list containing the labels for each task in the project.
        Each entry in the outermost list corresponds to a trainable task in the project.

        :param include_empty: True to include empty labels in the output, False to
            exclude them
        :return:
        """
        return [
            [attr.asdict(label) for label in task_labels]
            for task_labels in self.pipeline.get_labels_per_task(
                include_empty=include_empty
            )
        ]

    def get_parameters(self) -> Dict[str, Union[str, List[str]]]:
        """
        Return the parameters used to create the project.

        :return: Dictionary containing the keys `project_name`, `project_type` and
            `labels`
        """
        parameter_dict = {
            "project_name": self.name,
            "project_type": self.project_type,
            "labels": self.get_labels_per_task(include_empty=False),
        }
        remove_null_fields(parameter_dict)
        return parameter_dict

    def get_all_labels(self) -> List[Label]:
        """
        Return a list of all labels in the project.

        :return: List of all labels in the project
        """
        return self.pipeline.get_all_labels()

    @property
    def overview(self) -> str:
        """
        Return a string that shows an overview of the project. This still shows all
        the detailed information of the project. If less details are required, please
        use the `summary` property

        :return: String holding an overview of the project
        """
        deidentified = copy.deepcopy(self)
        deidentified.deidentify()
        overview_dict = deidentified.to_dict()
        remove_null_fields(overview_dict)
        return pformat(overview_dict)

    @property
    def summary(self) -> str:
        """
        Return a string that gives a very brief summary of the project. This is the
        least detailed representation of the project, if more details are required
        please use the `overview` property

        :return: String holding a brief summary of the project
        """
        summary_str = f"Project: {self.name}\n"
        for task_index, (task, labels) in enumerate(
            zip(self.get_trainable_tasks(), self.pipeline.get_labels_per_task())
        ):
            summary_str += f"  Task {task_index+1}: {task.title}\n"
            if labels is not None:
                summary_str += f"    Labels: {[label.name for label in labels]}\n"
        return summary_str

    @property
    def training_dataset(self) -> Dataset:
        """
        Return the training dataset for the project.

        :return: Training dataset for the project
        """
        return [dataset for dataset in self.datasets if dataset.use_for_training][0]



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
from typing import List

import numpy as np

from geti_sdk.data_models import Prediction
from geti_sdk.deployment.inference_hook_interfaces import PostInferenceTrigger


class LabelTrigger(PostInferenceTrigger):
    """
    Post inference trigger based on the presence of certain label(s) in the
    predictions for an image.
    If the specified label or labels are in the prediction results for an inferred
    image, the trigger is activated.

    :param label_names: List of label names to search for. If any of
    :param mode: Operation mode to use when evaluating the labels. Supported modes are:
        - `OR` (the default) -> If **any** of the labels in `label_names` if found in
            the prediction, the trigger activates.
        - `AND` -> If and only if **all** of the labels in `label_names` are
            predicted within the image, the trigger activates
    """

    def __init__(self, label_names: List[str], mode: str = "OR"):
        self.label_names = set(label_names)
        self.mode = mode

        # LabelTrigger will return a score of 1 if label is found, so we can use the
        # default threshold defined in the super class
        super().__init__()
        self._repr_info_ = f"label_names={label_names}"
        self._repr_info_ += f", mode={mode}"

    def __call__(self, image: np.ndarray, prediction: Prediction) -> float:
        """
        Compute a trigger score for the `image` and corresponding `prediction`.

        :param image: Numpy array representing an image
        :param prediction: Prediction object corresponding to the inference result
            for the image.
        :return: Float representing the score for the input
        """
        predicted_labels = set()
        for label in prediction.get_labels():
            predicted_labels.add(label.name)
        if self.mode == "AND":
            return float(self.label_names.issubset(predicted_labels))
        else:  # mode == "OR"
            return float(not self.label_names.isdisjoint(predicted_labels))



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import time

from geti_sdk import Geti
from geti_sdk.data_models import Project
from geti_sdk.data_models.enums import JobState
from geti_sdk.rest_clients import PredictionClient, TrainingClient


def ensure_project_is_trained(geti: Geti, project: Project) -> bool:
    """
    Ensure that the `project` has a trained model for each task.

    If no trained model is found for any of the tasks, the function will attempt to
    start training for that task. It will then await the completion of the training job.

    This method returns True if all tasks in the project have a trained model
    available, and the project is therefore ready to make predictions.

    :param geti: Geti instance pointing to the GETi server
    :param project: Project object, representing the project in GETi
    :return: True if the project is trained and ready to make predictions, False
        otherwise
    """
    prediction_client = PredictionClient(
        session=geti.session, workspace_id=geti.workspace_id, project=project
    )
    if prediction_client.ready_to_predict:
        print(f"\nProject '{project.name}' is ready to predict.\n")
        return True

    print(
        f"\nProject '{project.name}' is not ready for prediction yet, awaiting model "
        f"training completion.\n"
    )
    training_client = TrainingClient(
        session=geti.session, workspace_id=geti.workspace_id, project=project
    )
    # If there are no jobs running for the project, we launch them
    jobs = training_client.get_jobs(project_only=True)
    running_jobs = [job for job in jobs if job.state == JobState.RUNNING]
    tasks = project.get_trainable_tasks()

    new_jobs = []
    if len(running_jobs) != len(tasks):
        for task in project.get_trainable_tasks():
            new_jobs.append(training_client.train_task(task))

    # Monitor job progress to ensure training finishes
    training_client.monitor_jobs(running_jobs + new_jobs)

    tries = 20
    while not prediction_client.ready_to_predict and tries > 0:
        time.sleep(1)
        tries -= 1

    if prediction_client.ready_to_predict:
        print(f"\nProject '{project.name}' is ready to predict.\n")
        prediction_ready = True
    else:
        print(
            f"\nAll jobs completed, yet project '{project.name}' is still not ready "
            f"to predict. This is likely due to an error in the training job.\n"
        )
        prediction_ready = False
    return prediction_ready



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import hashlib
import logging
import os
import shutil
import zipfile
from typing import Dict, Optional

import requests
from tqdm import tqdm


def get_proxies(url: str = "", verify_cert: bool = True) -> Dict[str, str]:
    """
    Determine whether or not to use proxies to attempt to reach a certain url.

    :param url: URL that should be resolved
    :param verify_cert: False to disable SSL certificate validation
    :return:
    """
    logging.info(f"Connecting to url {url}...")
    proxies: Dict[str, str] = {}
    timeout = 10
    try:
        requests.head(url=url, proxies=proxies, timeout=timeout, verify=verify_cert)
        return proxies
    except requests.exceptions.ConnectionError:
        logging.info("Unable to reach URL, attempting to connect via proxy...")
    proxies = {
        "http": "http://proxy-mu.intel.com:911",
        "https": "http://proxy-mu.intel.com:912",
    }
    try:
        requests.head(url=url, proxies=proxies, verify=verify_cert, timeout=timeout)
        logging.info("Connection succeeded.")
    except requests.exceptions.ConnectionError as error:
        raise ValueError(
            "Unable to resolve URL with any proxy settings, please try to turn off "
            "your VPN connection before attempting again."
        ) from error
    return proxies


def download_file(
    url: str,
    target_folder: Optional[str],
    check_valid_archive: bool = False,
    verify_cert: bool = True,
    timeout: int = 1800,
) -> str:
    """
    Download a file from `url` to a folder on local disk `target_folder`.

    NOTE: If a file with the same name as the file to be downloaded already exists in
        `target_folder`, this function will not download anything. If
        `check_valid_archive` is True, this function not only checks if the target
        file exists but also if it is a valid .zip archive.

    :param url:
    :param target_folder:
    :param check_valid_archive: Check if the target file is a valid zip archive
    :param verify_cert: False to disable SSL certificate validation
    :param timeout: Time (in seconds) after which the download will time out
    :return: path to the downloaded file
    """
    filename = url.split("/")[-1]
    if target_folder is None:
        target_folder = "data"
    path_to_file = os.path.join(target_folder, filename)
    valid_file_exists = False
    if os.path.exists(path_to_file) and os.path.isfile(path_to_file):
        valid_file_exists = True
        if check_valid_archive:
            if not zipfile.is_zipfile(path_to_file):
                logging.info(
                    f"File {filename} exists at {path_to_file}, but is is not a valid "
                    f"archive. Overwriting the existing file."
                )
                try:
                    shutil.rmtree(path_to_file)
                except NotADirectoryError:
                    os.remove(path_to_file)
                valid_file_exists = False
    if valid_file_exists:
        logging.info(
            f"File {filename} exists at {path_to_file}. No new data was downloaded."
        )
        return path_to_file

    proxies = get_proxies(url, verify_cert=verify_cert)
    logging.info(f"Downloading {filename}...")
    with requests.get(
        url, stream=True, proxies=proxies, verify=verify_cert, timeout=timeout
    ) as r:
        if r.status_code != 200:
            r.raise_for_status()
            raise RuntimeError(
                f"Request to {url} failed, returned status code {r.status_code}"
            )
        file_size = int(r.headers.get("Content-Length", 0))
        with tqdm.wrapattr(r.raw, "read", total=file_size, desc="") as r_raw:
            with open(path_to_file, "wb") as f:
                shutil.copyfileobj(r_raw, f)
    logging.info("Download complete.")
    return path_to_file


def validate_hash(file_path: str, expected_hash: str) -> None:
    """
    Verify that hash matches the calculated hash of the file.

    :param file_path: Path to file.
    :param expected_hash: Expected hash of the file.
    """
    with open(file_path, "rb") as hash_file:
        downloaded_hash = hashlib.sha3_512(hash_file.read()).hexdigest()
    if downloaded_hash != expected_hash:
        raise ValueError(
            f"Downloaded file {file_path} does not match the required hash."
        )


def set_directory_permissions(
    target_directory: str, file_permissions=0o660, dir_permissions=0o770
) -> None:
    """
    Set read, write, execute permissions for user and user_group on a directory tree.

    NOTE: Please use this method with caution, on temporary directories only

    :param target_directory: path to the directory for which to set the permissions
    :param file_permissions: Permissions to apply to all files found in the directory
        tree
    :param dir_permissions: Permissions to apply to all directories found in the
        directory tree
    """
    os.chmod(target_directory, dir_permissions)  # nosec: B103
    for root, dirs, files in os.walk(target_directory):
        for file_name in files:
            file_path = os.path.join(root, file_name)
            os.chmod(file_path, file_permissions)
        for dir_name in dirs:
            dir_path = os.path.join(root, dir_name)
            os.chmod(dir_path, dir_permissions)  # nosec: B103



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import json
import logging
import os
import shutil
import tempfile
import time
import zipfile
from typing import Any, Dict, List, Optional, Tuple, Union

import attr
import numpy as np
from model_api.adapters import OpenvinoAdapter, OVMSAdapter
from model_api.models import Model as model_api_Model
from model_api.tilers import DetectionTiler, InstanceSegmentationTiler, Tiler
from openvino.runtime import Core
from packaging.version import Version

from geti_sdk.data_models import OptimizedModel, Project, TaskConfiguration
from geti_sdk.data_models.enums.domain import Domain
from geti_sdk.data_models.label import Label
from geti_sdk.data_models.label_group import LabelGroup, LabelGroupType
from geti_sdk.data_models.label_schema import LabelSchema
from geti_sdk.data_models.predictions import Prediction, ResultMedium
from geti_sdk.deployment.predictions_postprocessing.results_converter.results_to_prediction_converter import (
    ConverterFactory,
    InferenceResultsToPredictionConverter,
)
from geti_sdk.http_session import GetiSession
from geti_sdk.rest_converters import ConfigurationRESTConverter, ModelRESTConverter

from .utils import (
    generate_ovms_model_address,
    generate_ovms_model_name,
    get_package_version_from_requirements,
    rgb_to_hex,
    target_device_is_ovms,
)

MODEL_DIR_NAME = "model"
PYTHON_DIR_NAME = "python"
REQUIREMENTS_FILE_NAME = "requirements.txt"

LABELS_CONFIG_KEY = "labels"
LABEL_TREE_KEY = "label_tree"
LABEL_GROUPS_KEY = "label_groups"
ALL_LABELS_KEY = "all_labels"

SALIENCY_KEY = "saliency_map"
ANOMALY_SALIENCY_KEY = "anomaly_map"
SEGMENTATION_SALIENCY_KEY = "soft_prediction"
FEATURE_VECTOR_KEY = "feature_vector"

TILER_MAPPING = {
    Domain.DETECTION: DetectionTiler,
    Domain.INSTANCE_SEGMENTATION: InstanceSegmentationTiler,
}

OVMS_TIMEOUT = 10  # Max time to wait for OVMS models to become available


@attr.define
class DeployedModel(OptimizedModel):
    """
    Representation of an Intel® Geti™ model that has been deployed for inference. It
    can be loaded onto a device to generate predictions.
    """

    hyper_parameters: Optional[TaskConfiguration] = attr.field(
        kw_only=True, repr=False, default=None
    )

    def __attrs_post_init__(self):
        """
        Initialize private attributes
        """
        super().__attrs_post_init__()
        self._model_data_path: Optional[str] = None
        self._model_python_path: Optional[str] = None
        self._needs_tempdir_deletion: bool = False
        self._tempdir_path: Optional[str] = None
        self._label_schema: Optional[LabelSchema] = None

        # Attributes related to model explainability
        self._saliency_key: Optional[str] = None
        self._saliency_location: Optional[str] = None
        self._feature_vector_key: Optional[str] = None
        self._feature_vector_location: Optional[str] = None

        self._converter: Optional[InferenceResultsToPredictionConverter] = None
        self._tiling_enabled: bool = False
        self._tiler: Optional[Tiler] = None

    @property
    def model_data_path(self) -> str:
        """
        Return the path to the raw model data

        :return: path to the directory containing the raw model data
        """
        if self._model_data_path is None:
            raise ValueError(
                "Model data path has not been set yet, location of binary model data "
                "is unknown."
            )
        return self._model_data_path

    def get_data(self, source: Union[str, os.PathLike, GetiSession]):
        """
        Load the model weights from a data source. The `source` can be one of the
        following:

          1. The Intel® Geti™ platform (if an GetiSession instance is passed). In this
            case the weights will be downloaded, and extracted to a temporary directory
          2. A zip file on local disk, in this case the weights will be extracted to a
             temporary directory
          3. A folder on local disk containing the .xml and .bin file for the model

        :param source: Data source to load the weights from
        """
        if isinstance(source, (os.PathLike, str)):
            if os.path.isfile(source) and os.path.splitext(source)[1] == ".zip":
                # Extract zipfile into temporary directory
                if self._model_data_path is None:
                    temp_dir = tempfile.mkdtemp()
                    self._needs_tempdir_deletion = True
                    self._tempdir_path = temp_dir
                else:
                    temp_dir = self._model_data_path

                with zipfile.ZipFile(source, "r") as zipped_source_model:
                    zipped_source_model.extractall(temp_dir)

                # _model_data_path contains the model structure and weights
                # _model_python_path contains the custom model wrappers
                self._model_data_path = os.path.join(temp_dir, MODEL_DIR_NAME)
                self._model_python_path = os.path.join(temp_dir, PYTHON_DIR_NAME)

                self.get_data(temp_dir)

            elif os.path.isdir(source):
                source_contents = os.listdir(source)
                if MODEL_DIR_NAME in source_contents:
                    model_dir = os.path.join(source, MODEL_DIR_NAME)
                else:
                    model_dir = source
                model_dir_contents = os.listdir(model_dir)
                if (
                    "model.bin" in model_dir_contents
                    and "model.xml" in model_dir_contents
                ):
                    self._model_data_path = model_dir
                else:
                    raise ValueError(
                        f"Unable to load model data from path '{model_dir}'. Model "
                        f"file 'model.xml' and weights file 'model.bin' were not found "
                        f"at the path specified. "
                    )

                self._model_python_path = os.path.join(source, PYTHON_DIR_NAME)

            # A model is being loaded from disk, check if it is a legacy model
            # We support OTX models starting from version 1.5.0
            otx_version = get_package_version_from_requirements(
                requirements_path=os.path.join(
                    self._model_python_path, REQUIREMENTS_FILE_NAME
                ),
                package_name="otx",
            )
            if otx_version:  # Empty string if package not found
                if Version(otx_version) < Version("1.5.0"):
                    raise ValueError(
                        "\n"
                        "This deployment model is not compatible with the current SDK. Proposed solutions:\n"
                        "1. Please deploy a model using GETi Platform version 2.0.0 or higher.\n"
                        "2. Downgrade to a compatible GETi-SDK version to continue using this model.\n\n"
                    )

        elif isinstance(source, GetiSession):
            if self.base_url is None:
                raise ValueError(
                    f"Insufficient data to retrieve data for model {self}. Please set "
                    f"a base_url for the model first."
                )
            response = source.get_rest_response(
                url=self.base_url + "/export", method="GET", contenttype="zip"
            )
            filename = f"{self.name}_{self.optimization_type}_optimized.zip"
            model_dir = tempfile.mkdtemp()
            model_filepath = os.path.join(model_dir, filename)
            with open(model_filepath, "wb") as f:
                f.write(response.content)
            self._model_data_path = model_dir
            self._needs_tempdir_deletion = True
            self._tempdir_path = model_dir
            self.get_data(source=model_filepath)

    def __del__(self):
        """
        Clean up the temporary directory created to store the model data (if any). This
        method is called when the OptimizedModel object is deleted.
        """
        if self._needs_tempdir_deletion:
            if self._tempdir_path is not None and os.path.exists(self._tempdir_path):
                shutil.rmtree(self._tempdir_path)

    def load_inference_model(
        self,
        device: str = "CPU",
        configuration: Optional[Dict[str, Any]] = None,
        project: Optional[Project] = None,
    ) -> None:
        """
        Load the actual model weights to a specified device.

        :param device: Device (CPU or GPU) to load the model to. Defaults to 'CPU'
        :param configuration: Optional dictionary holding additional configuration
            parameters for the model
        :param project: Optional project to which the model belongs.
            This is only used when the model is run on OVMS, in that case the
            project is needed to identify the correct model
        :return: OpenVino inference engine model that can be used to make predictions
            on images
        """
        core = Core()
        if not target_device_is_ovms(device=device):
            # Run the model locally
            model_adapter = OpenvinoAdapter(
                core,
                model=os.path.join(self._model_data_path, "model.xml"),
                weights_path=os.path.join(self._model_data_path, "model.bin"),
                device=device,
                plugin_config=None,
                max_num_requests=1,
            )
        else:
            # Connect to an OpenVINO model server instance
            model_name = generate_ovms_model_name(project=project, model=self)
            model_address = generate_ovms_model_address(
                ovms_address=device, model_name=model_name
            )

            ovms_connected = False
            ovms_error: Optional[BaseException] = None
            t_start = time.time()
            while not ovms_connected and time.time() - t_start < OVMS_TIMEOUT:
                # If OVMS has just started, model needs some time to initialize
                try:
                    model_adapter = OVMSAdapter(model_address)
                    ovms_connected = True
                except RuntimeError as error:
                    time.sleep(0.5)
                    ovms_error = error
            if not ovms_connected:
                if ovms_error is not None:
                    raise RuntimeError("Unable to connect to OVMS") from ovms_error
                else:
                    raise RuntimeError(
                        "Unknown error encountered while connecting to OVMS"
                    )

        # Load model configuration
        config_path = os.path.join(self._model_data_path, "config.json")
        if not os.path.isfile(config_path):
            raise ValueError(
                f"Missing configuration file `config.json` for deployed model `{self}`,"
                f" unable to load inference model."
            )
        with open(config_path, "r") as config_file:
            configuration_json = json.load(config_file)
        model_type = configuration_json.get("type_of_model")

        # Update model parameters
        parameters = configuration_json.get("model_parameters")
        if configuration is not None:
            configuration.update(parameters)
        else:
            configuration = parameters

        # Parse label schema
        label_dictionary = configuration_json.get("model_parameters").pop(
            LABELS_CONFIG_KEY, None
        )
        self._parse_label_schema_from_dict(label_dictionary)

        # Load a Results-to-Prediction converter
        self._converter = ConverterFactory.create_converter(
            self.label_schema, configuration
        )

        model = model_api_Model.create_model(
            model=model_adapter,
            model_type=model_type,
            preload=False,
            core=core,
            configuration=configuration,
        )
        self._inference_model = model

        # Extract tiling parameters, if applicable
        tiling_parameters = configuration_json.get("tiling_parameters", None)

        enable_tiling = False
        if tiling_parameters is not None:
            enable_tiling = tiling_parameters.get("enable_tiling", False)
            if isinstance(enable_tiling, dict):
                enable_tiling = enable_tiling.get("value", False)

        if enable_tiling:
            logging.info("Tiling is enabled for this model, initializing Tiler")
            tiler_type = TILER_MAPPING.get(self._converter.domain, None)
            if tiler_type is None:
                raise ValueError(
                    f"Tiling is not supported for domain {self._converter.domain}"
                )
            self._tiler = tiler_type(model=model, execution_mode="sync")
            self._tiling_enabled = True

        # TODO: This is a workaround to fix the issue that causes the output blob name
        #  to be unset. Remove this once it has been fixed on ModelAPI side
        output_names = list(self._inference_model.outputs.keys())
        if hasattr(self._inference_model, "output_blob_name"):
            if not self._inference_model.output_blob_name:
                self._inference_model.output_blob_name = {
                    name: name for name in output_names
                }

        # Force reload model to account for any postprocessing changes that may have
        # been applied while creating the ModelAPI wrapper
        logging.info(
            f"Inference model wrapper initialized, force reloading model on device "
            f"`{device}` to finalize inference model initialization process."
        )
        self._inference_model.load(force=True)

    @classmethod
    def from_model_and_hypers(
        cls, model: OptimizedModel, hyper_parameters: Optional[TaskConfiguration] = None
    ) -> "DeployedModel":
        """
        Create a DeployedModel instance out of an OptimizedModel and it's
        corresponding set of hyper parameters.

        :param model: OptimizedModel to convert to a DeployedModel
        :param hyper_parameters: TaskConfiguration instance containing the hyper
            parameters for the model
        :return: DeployedModel instance
        """
        model_dict = model.to_dict()
        model_dict.update({"hyper_parameters": hyper_parameters})
        deployed_model = cls(**model_dict)
        try:
            model_group_id = model.model_group_id
            base_url = model.base_url
            deployed_model.model_group_id = model_group_id
            deployed_model.base_url = base_url
        except ValueError:
            pass
        return deployed_model

    @classmethod
    def from_folder(cls, path_to_folder: Union[str, os.PathLike]) -> "DeployedModel":
        """
        Create a DeployedModel instance from a folder containing the model data.

        :param path_to_folder: Path to the folder that holds the model data
        :return: DeployedModel instance
        """
        config_filepath = os.path.join(path_to_folder, "hyper_parameters.json")
        if os.path.isfile(config_filepath):
            with open(config_filepath, "r") as config_file:
                config_dict = json.load(config_file)
            hparams = ConfigurationRESTConverter.task_configuration_from_dict(
                config_dict
            )
        else:
            hparams = None
        model_detail_path = os.path.join(path_to_folder, "model.json")
        with open(model_detail_path, "r") as model_detail_file:
            model_detail_dict = json.load(model_detail_file)
        model = ModelRESTConverter.optimized_model_from_dict(model_detail_dict)
        deployed_model = cls.from_model_and_hypers(
            model=model, hyper_parameters=hparams
        )
        deployed_model.get_data(source=path_to_folder)
        return deployed_model

    def save(self, path_to_folder: Union[str, os.PathLike]) -> bool:
        """
        Save the DeployedModel instance to the designated folder.

        :param path_to_folder: Path to the folder to save the model to
        :return: True if the model was saved successfully, False otherwise
        """
        if self._model_data_path is None:
            raise ValueError(
                "No model definition and model weights data was found for {self}, "
                "unable to save model."
            )
        os.makedirs(path_to_folder, exist_ok=True, mode=0o770)

        new_model_data_path = os.path.join(path_to_folder, MODEL_DIR_NAME)
        new_model_python_path = os.path.join(path_to_folder, PYTHON_DIR_NAME)

        if (
            new_model_python_path == self._model_python_path
            or new_model_data_path == self._model_data_path
        ):
            logging.warning(
                f"Model '{self.name}' already exist in target path {path_to_folder}, "
                f"please save to a different location."
            )
            return False

        shutil.copytree(
            src=self._model_data_path,
            dst=new_model_data_path,
            dirs_exist_ok=True,
        )
        if self._model_python_path is not None:
            shutil.copytree(
                src=self._model_python_path,
                dst=new_model_python_path,
                dirs_exist_ok=True,
            )
            self._model_python_path = new_model_python_path

        self._model_data_path = new_model_data_path

        config_dict = ConfigurationRESTConverter.configuration_to_minimal_dict(
            self.hyper_parameters
        )
        config_filepath = os.path.join(path_to_folder, "hyper_parameters.json")
        with open(config_filepath, "w") as config_file:
            json.dump(config_dict, config_file, indent=4)

        model_detail_dict = self.to_dict()
        model_detail_dict.pop("hyper_parameters")
        model_detail_path = os.path.join(path_to_folder, "model.json")
        with open(model_detail_path, "w") as model_detail_file:
            json.dump(model_detail_dict, model_detail_file, indent=4)
        return True

    def _preprocess(
        self, image: np.ndarray
    ) -> Tuple[Dict[str, np.ndarray], Dict[str, Tuple[int, int, int]]]:
        """
        Preprocess an image for inference.

        :param image: Numpy array containing pixel data. The image is expected to have
            dimensions [height x width x channels], with the channels in RGB order
        :return: Dictionary holding the preprocessing result, the original shape of
            the image and the shape after preprocessing
        """
        return self._inference_model.preprocess(image)

    def _postprocess(
        self,
        inference_results: Dict[str, np.ndarray],
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Union[np.ndarray, List[Tuple[int, float]], Any]:
        """
        Postprocess the model outputs.

        :param inference_results: Dictionary holding the results of inference
        :param metadata: Dictionary holding metadata
        :return: Postprocessed inference results. The exact format depends on the
            type of model that is loaded:

            * For segmentation models, it will be a numpy array holding the output mask
            * For classification models, it will be a list of tuples holding the
                output class index and class probability
            * For detection models, it will be an instance of
                `openvino.model_zoo.model_api.models.utils.Detection`, holding the
                bounding box output
        """
        return self._inference_model.postprocess(inference_results, metadata)

    def _postprocess_explain_outputs(
        self,
        inference_results: Dict[str, np.ndarray],
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Postprocess the model outputs to obtain saliency maps, feature vectors and
        active scores.

        :param inference_results: Dictionary holding the results of inference
        :param metadata: Dictionary holding metadata
        :return: Tuple containing postprocessed outputs, formatted as follows:
            - Numpy array containing the saliency map
            - Numpy array containing the feature vector
        """
        if self._saliency_location is None and self._feature_vector_location is None:
            # When running postprocessing for the first time, we need to determine the
            # key and location of the saliency map and feature vector.
            if hasattr(self._inference_model, "postprocess_aux_outputs"):
                (
                    _,
                    saliency_map,
                    repr_vector,
                    _,
                ) = self._inference_model.postprocess_aux_outputs(
                    inference_results, metadata
                )
                self._saliency_location = "aux"
                self._feature_vector_location = "aux"
            else:
                # Check all possible saliency map keys in outputs and metadata
                if SALIENCY_KEY in inference_results.keys():
                    saliency_map = inference_results[SALIENCY_KEY]
                    self._saliency_location = "output"
                    self._saliency_key = SALIENCY_KEY
                elif SALIENCY_KEY in metadata.keys():
                    saliency_map = metadata[SALIENCY_KEY]
                    self._saliency_location = "meta"
                    self._saliency_key = SALIENCY_KEY
                elif ANOMALY_SALIENCY_KEY in metadata.keys():
                    saliency_map = metadata[ANOMALY_SALIENCY_KEY]
                    self._saliency_location = "meta"
                    self._saliency_key = ANOMALY_SALIENCY_KEY
                elif SEGMENTATION_SALIENCY_KEY in metadata.keys():
                    saliency_map = metadata[SEGMENTATION_SALIENCY_KEY]
                    self._saliency_location = "meta"
                    self._saliency_key = SEGMENTATION_SALIENCY_KEY
                else:
                    logging.warning("No saliency map found in model output")
                    saliency_map = None

                # Check all possible feature vector keys in outputs and metadata
                if FEATURE_VECTOR_KEY in inference_results.keys():
                    repr_vector = inference_results[FEATURE_VECTOR_KEY]
                    self._feature_vector_location = "output"
                    self._feature_vector_key = FEATURE_VECTOR_KEY
                elif FEATURE_VECTOR_KEY in metadata.keys():
                    repr_vector = metadata[FEATURE_VECTOR_KEY]
                    self._feature_vector_location = "meta"
                    self._feature_vector_key = FEATURE_VECTOR_KEY
                else:
                    logging.warning("No feature vector found in model output")
                    repr_vector = None
        else:
            # If location of feature vector and saliency map are already known, we can
            # use them directly
            if self._saliency_location == "aux":
                (
                    _,
                    saliency_map,
                    repr_vector,
                    _,
                ) = self._inference_model.postprocess_aux_outputs(
                    inference_results, metadata
                )
                return saliency_map, repr_vector
            elif self._saliency_location == "meta":
                saliency_map = metadata[self._saliency_key]
            elif self._saliency_location == "output":
                saliency_map = inference_results[self._saliency_key]
            else:
                logging.warning("No saliency map found in model output")
                saliency_map = None
            if self._feature_vector_location == "meta":
                repr_vector = metadata[self._feature_vector_key]
            elif self._feature_vector_location == "output":
                repr_vector = inference_results[self._feature_vector_key]

            else:
                logging.warning("No feature vector found in model output")
                repr_vector = None
        return saliency_map, repr_vector

    def infer(self, image: np.ndarray, explain: bool = False) -> Prediction:
        """
        Run inference on an already preprocessed image.

        :param preprocessed_image: Dictionary holding the preprocessing results for an
            image
        :return: Dictionary containing the model outputs
        """
        if not self._tiling_enabled:
            preprocessed_image, metadata = self._preprocess(image)
            # metadata is a dict with keys 'original_shape' and 'resized_shape'
            inference_results: Dict[str, np.ndarray] = self._inference_model.infer_sync(
                preprocessed_image
            )
            postprocessing_results = self._postprocess(
                inference_results, metadata=metadata
            )
        else:
            postprocessing_results = self._tiler(image)

        prediction = self._converter.convert_to_prediction(
            postprocessing_results, image_shape=image.shape
        )

        # Add optional explainability outputs
        if explain:
            if not self._tiling_enabled:
                saliency_map, repr_vector = self._postprocess_explain_outputs(
                    inference_results=inference_results, metadata=metadata
                )
            else:
                repr_vector = postprocessing_results.feature_vector
                saliency_map = postprocessing_results.saliency_map

            prediction.feature_vector = repr_vector
            result_medium = ResultMedium(name="saliency map", type="saliency map")
            result_medium.data = saliency_map
            prediction.maps = [result_medium]

        return prediction

    @property
    def label_schema(self) -> LabelSchema:
        """
        Return the LabelSchema for the model.

        This requires the inference model to be loaded, getting this property while
        inference models are not loaded will raise a ValueError

        :return: LabelSchema containing the SDK label schema for the model
        """
        if self._label_schema is None:
            raise ValueError(
                "Inference model is not loaded, unable to retrieve label schema. "
                "Please load inference model first."
            )
        return self._label_schema

    def _parse_label_schema_from_dict(
        self, label_schema_dict: Optional[Dict[str, Union[dict, List[dict]]]] = None
    ) -> None:
        """
        Parse the dictionary contained in the model `config.json` file, and
        generate an  LabelSchema from it.

        :param label_schema_dict: Dictionary containing the label schema information
            to parse
        """
        label_groups_list = label_schema_dict[LABEL_GROUPS_KEY]
        labels_dict = label_schema_dict[ALL_LABELS_KEY]
        for key, value in labels_dict.items():
            color_tuple = tuple(
                int(value["color"][key]) for key in ["red", "green", "blue"]
            )
            label_entity = Label(
                name=value["name"],
                color=rgb_to_hex(color_tuple),
                group=None,
                is_empty=value.get("is_empty", False),
                hotkey=value["hotkey"],
                domain=Domain[value["domain"]],
                id=value["_id"],
                is_anomalous=value.get("is_anomalous", False),
            )
            labels_dict[key] = label_entity
        label_groups: List[LabelGroup] = []
        for group_dict in label_groups_list:
            labels: List[Label] = [
                labels_dict[label_id] for label_id in group_dict["label_ids"]
            ]
            label_groups.append(
                LabelGroup(
                    id=group_dict["_id"],
                    name=group_dict["name"],
                    group_type=LabelGroupType[group_dict["relation_type"]],
                    labels=labels,
                )
            )
        self._label_schema = LabelSchema(label_groups=label_groups)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import logging
from typing import Optional

from geti_sdk import Geti
from geti_sdk.annotation_readers import DirectoryTreeAnnotationReader
from geti_sdk.data_models import Project
from geti_sdk.demos.data_helpers import get_mvtec_dataset
from geti_sdk.rest_clients import AnnotationClient, ImageClient, ProjectClient

from .utils import ensure_project_is_trained


def create_anomaly_classification_demo_project(
    geti: Geti,
    n_images: int,
    n_annotations: int = -1,
    project_name: str = "Anomaly classification demo",
    dataset_path: Optional[str] = None,
) -> Project:
    """
    Create a demo project of type 'anomaly_classification', based off the MVTec
    anomaly detection dataset.

    This method creates a project with a single 'Anomaly classification' task.

    :param geti: Geti instance, representing the GETi server on which
        the project should be created.
    :param n_images: Number of images that should be uploaded. Pass -1 to upload all
        available images in the dataset for the given labels
    :param n_annotations: Number of images that should be annotated. Pass -1 to
        upload annotations for all images.
    :param project_name: Name of the project to create.
        Defaults to 'Anomaly classification demo'
    :param dataset_path: Path to the dataset to use as data source. Defaults to
        the 'data' directory in the top level folder of the geti_sdk package. If
        the dataset is not found in the target folder, this method will attempt to
        download it from the internet.
    :return: Project object, holding detailed information about the project that was
        created on the Intel® Geti™ server.
    """
    project_client = ProjectClient(session=geti.session, workspace_id=geti.workspace_id)
    data_path = get_mvtec_dataset(dataset_path)
    logging.info(" ------- Creating anomaly classification project --------------- ")

    # Create annotation reader
    annotation_reader = DirectoryTreeAnnotationReader(
        base_data_folder=data_path, subset_folder_names=["train", "test"]
    )
    annotation_reader.group_labels(
        labels_to_group=["damaged_case", "cut_lead", "misplaced", "bent_lead"],
        group_name="Anomalous",
    )
    annotation_reader.group_labels(labels_to_group=["good"], group_name="Normal")

    # Create the project
    project = project_client.create_project(
        project_name=project_name,
        project_type="anomaly_classification",
        labels=[[]],
    )

    # Upload images to the project
    data_filepaths = annotation_reader.get_data_filenames()
    image_client = ImageClient(
        session=geti.session, workspace_id=geti.workspace_id, project=project
    )
    images = image_client.upload_from_list(
        path_to_folder=data_path,
        image_names=data_filepaths,
        extension_included=False,
        image_names_as_full_paths=True,
        n_images=n_images,
    )
    annotation_client = AnnotationClient(
        session=geti.session,
        workspace_id=geti.workspace_id,
        project=project,
        annotation_reader=annotation_reader,
    )

    # The annotation_reader is used to upload annotations for all images
    for image, full_path in zip(images, data_filepaths):
        image.name = full_path

    if n_annotations == -1:
        n_annotations = len(images) - 1
    annotation_client.upload_annotations_for_images(images=images[:n_annotations])

    return project


def ensure_trained_anomaly_project(
    geti: Geti, project_name: str = "Transistor anomaly classification"
):
    """
    Check whether the project named `project_name` exists on the server, and create it
    if it not.

    If the project does not exist, this method will create an anomaly classification
    project based on the MVTec AD `transistor` dataset.

    :param geti: Geti instance pointing to the Intel® Geti™ server
    :param project_name: Name of the project to look for or create
    :return: Project object representing the project on the server
    """
    project_client = ProjectClient(session=geti.session, workspace_id=geti.workspace_id)
    project = project_client.get_project_by_name(project_name)
    if project is None:
        logging.info(f"Project '{project_name}' does not exist yet, creating it...")

        project = create_anomaly_classification_demo_project(
            geti=geti, n_images=-1, project_name=project_name
        )
        logging.info(
            f"Project `{project_name}` of type `anomaly_classification` was created on "
            f"host `{geti.session.config.host}`."
        )

    ensure_project_is_trained(geti, project)
    return project



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
import time

from geti_sdk.rest_clients import ProjectClient, TrainingClient


def force_delete_project(project_name: str, project_client: ProjectClient) -> None:
    """
    Deletes the project named 'project_name'. If any jobs are running for the
    project, this finalizer cancels them.

    :param project_name: Name of project to delete
    :param project_client: ProjectClient to use for project deletion
    """
    try:
        project_client.delete_project(project=project_name, requires_confirmation=False)
    except TypeError:
        logging.warning(
            f"Project {project_name} was not found on the server, it was most "
            f"likely already deleted."
        )
    except ValueError:
        logging.error(
            f"Unable to delete project '{project_name}' from the server, it "
            f"is most likely locked for deletion due to an operation/training "
            f"session that is in progress. "
            f"\n\n Attempting to cancel the job and re-try project deletion."
        )

        project = project_client.get_project_by_name(project_name)
        training_client = TrainingClient(
            workspace_id=project_client.workspace_id,
            session=project_client.session,
            project=project,
        )
        jobs = training_client.get_jobs(project_only=True)
        for job in jobs:
            job.cancel(project_client.session)
        time.sleep(1)
        try:
            project_client.delete_project(
                project=project_name, requires_confirmation=False
            )
        except ValueError as error:
            raise ValueError(
                f"Unable to force delete project {project_name}, due to: {error} "
            )



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Any, Dict, Type

from geti_sdk.data_models.containers.media_list import MediaTypeVar
from geti_sdk.utils import deserialize_dictionary


class MediaRESTConverter:
    """
    Class that handles conversion of Intel® Geti™ REST output for media entities to
    objects and vice versa.
    """

    @staticmethod
    def from_dict(
        input_dict: Dict[str, Any], media_type: Type[MediaTypeVar]
    ) -> MediaTypeVar:
        """
        Create an instance of type `media_type` representing a media entity on the
        Intel® Geti™ server from a dictionary returned by the GETi /media REST
        endpoints.

        :param input_dict: Dictionary representing the media entity
        :param media_type: Type of the media entity
        :return: Instance of type `media_type` containing the entity represented in
            the REST input in `input_dict`.
        """
        return deserialize_dictionary(input_dict, output_type=media_type)



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""The module implements helpers for drawing shapes."""

# pylint: disable=too-many-instance-attributes, too-many-arguments, too-many-locals
import abc
from typing import (
    Callable,
    Generic,
    List,
    NewType,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import cv2
import numpy as np

from geti_sdk.data_models.annotation_scene import AnnotationScene
from geti_sdk.data_models.annotations import Annotation
from geti_sdk.data_models.label import Label, ScoredLabel
from geti_sdk.data_models.predictions import Prediction
from geti_sdk.data_models.shapes import (
    Ellipse,
    Point,
    Polygon,
    Rectangle,
    RotatedRectangle,
    Shape,
)

CvTextSize = NewType("CvTextSize", Tuple[Tuple[int, int], int])

_Any = TypeVar("_Any")


class DrawerEntity(Generic[_Any]):
    """An interface to draw a shape of type ``T`` onto an image."""

    supported_types: Sequence[Type[Shape]] = []

    @abc.abstractmethod
    def draw(
        self, image: np.ndarray, entity: _Any, labels: List[ScoredLabel]
    ) -> np.ndarray:
        """
        Draw an entity to a given frame.

        :param image: The image to draw the entity on.
        :param entity: The entity to draw.
        :param labels: Labels of the shapes to draw
        :return: frame with shape drawn on it
        """
        raise NotImplementedError


class Helpers:
    """
    Contains variables which are used by all subclasses.

    Contains functions which help with generating coordinates, text and text scale.
    These functions are use by the DrawerEntity Classes when drawing to an image.
    """

    def __init__(self) -> None:
        # Same alpha value that the UI uses for Labels
        self.alpha_shape = 100 / 256
        self.alpha_labels = 153 / 256
        self.assumed_image_width_for_text_scale = (
            1500  # constant number for size of classification/counting overlay
        )
        self.top_margin = 0.07  # part of the top screen reserved for top left classification/counting overlay
        self.content_padding = 3
        self.top_left_box_thickness = 1
        self.content_margin = 2
        self.label_offset_box_shape = 10
        self.black = (0, 0, 0)
        self.white = (255, 255, 255)
        self.yellow = (255, 255, 0)

        self.cursor_pos = Point(0, 0)
        self.line_height = 0
        self.font = cv2.FONT_HERSHEY_DUPLEX

    @staticmethod
    def draw_transparent_rectangle(
        img: np.ndarray,
        x1: int,
        y1: int,
        x2: int,
        y2: int,
        color: Tuple[int, int, int],
        alpha: float,
    ) -> np.ndarray:
        """
        Draw a rectangle on an image.

        :param img: Image
        :param x1: Left side
        :param y1: Top side
        :param x2: Right side
        :param y2: Bottom side
        :param color: Color
        :param alpha: Alpha value between 0 and 1
        :return: Image with rectangle drawn on it
        """
        x1 = np.clip(x1, 0, img.shape[1] - 1)
        y1 = np.clip(y1, 0, img.shape[0] - 1)
        x2 = np.clip(x2 + 1, 0, img.shape[1] - 1)
        y2 = np.clip(y2 + 1, 0, img.shape[0] - 1)
        rect = img[y1:y2, x1:x2]
        rect[:] = (alpha * np.array(color))[np.newaxis, np.newaxis] + (1 - alpha) * rect
        return img

    def generate_text_scale(self, image: np.ndarray) -> float:
        """
        Calculate the scale of the text.

        :param image: Image to calculate the text scale for.
        :return: Scale for the text
        """
        return round(image.shape[1] / self.assumed_image_width_for_text_scale, 1)

    @staticmethod
    def generate_text_for_label(
        label: Union[Label, ScoredLabel], show_labels: bool, show_confidence: bool
    ) -> str:
        """
        Return a string representing a given label and its associated probability if label is a ScoredLabel.

        :param label: Label
        :param show_labels: Whether to render the labels above the shape
        :param show_confidence: Whether to render the confidence above the shape
        :return: Formatted string (e.g. `"Cat 58%"`)
        """
        text = ""
        if show_labels:
            text += label.name
        if show_confidence and isinstance(label, ScoredLabel):
            if len(text) > 0:
                text += " "
            text += f"{label.probability:.0%}"
        return text

    def generate_draw_command_for_labels(
        self,
        labels: Sequence[Union[Label, ScoredLabel]],
        image: np.ndarray,
        show_labels: bool,
        show_confidence: bool,
    ) -> Tuple[Callable[[np.ndarray], np.ndarray], int, int]:
        """
        Generate draw function and content width and height for labels.

        Generates a function which can be called to draw a list of labels onto an image relatively to the
        cursor position.
        The width and height of the content is also returned and can be determined to compute
        the best position for content before actually drawing it.

        :param labels: List of labels
        :param image: Image (used to compute font size)
        :param show_labels: Whether to show the label name
        :param show_confidence: Whether to show the confidence probability
        :return: A tuple containing the drawing function, the content width, and the content height
        """
        draw_commands = []
        content_width = 0
        content_height = 0

        # Loop through the list of labels and create a function which can be used to draw the label.
        n_labels = len(labels)
        for i, label in enumerate(labels):
            text = self.generate_text_for_label(label, show_labels, show_confidence)
            if i < n_labels - 1:
                text += " >"
            text_scale = self.generate_text_scale(image)
            thickness = int(text_scale / 2)
            color = label.color_tuple

            item_command, item_width, item_height = self.generate_draw_command_for_text(
                text, text_scale, thickness, color
            )

            draw_commands.append(item_command)

            content_width += item_width
            content_height = max(content_height, item_height)

        def draw_command(img: np.ndarray) -> np.ndarray:
            for command in draw_commands:
                img = command(img)
            return img

        return draw_command, content_width, content_height

    def generate_draw_command_for_text(
        self, text: str, text_scale: float, thickness: int, color: Tuple[int, int, int]
    ) -> Tuple[Callable[[np.ndarray], np.ndarray], int, int]:
        """
        Generate function to draw text on image relative to cursor position.

        Generate a function which can be called to draw the given text onto an image
        relatively to the cursor position.

        The width and height of the content is also returned and can be determined to compute
        the best position for content before actually drawing it.

        :param text: Text to draw
        :param text_scale: Font size
        :param thickness: Thickness of the text
        :param color: Color of the text
        :return: A tuple containing the drawing function, the content width, and the content height
        """
        padding = self.content_padding
        margin = self.content_margin

        (text_width, text_height), baseline = cv2.getTextSize(
            text, self.font, fontScale=text_scale, thickness=thickness
        )

        width = text_width + 2 * padding
        height = text_height + baseline + 2 * padding
        content_width = width + margin

        if (color[0] + color[1] + color[2]) / 3 > 200:
            text_color = self.black
        else:
            text_color = self.white

        def draw_command(img: np.ndarray) -> np.ndarray:
            cursor_pos = Point(int(self.cursor_pos.x), int(self.cursor_pos.y))
            self.draw_transparent_rectangle(
                img,
                int(cursor_pos.x),
                int(cursor_pos.y),
                int(cursor_pos.x + width),
                int(cursor_pos.y + height),
                color,
                self.alpha_labels,
            )

            img = cv2.putText(
                img=img,
                text=text,
                org=(
                    cursor_pos.x + padding,
                    cursor_pos.y + height - padding - baseline,
                ),
                fontFace=self.font,
                fontScale=text_scale,
                color=text_color,
                thickness=thickness,
                lineType=cv2.LINE_AA,
            )

            self.cursor_pos.x += content_width
            self.line_height = height

            return img

        return draw_command, content_width, height

    @staticmethod
    def draw_flagpole(
        image: np.ndarray,
        flagpole_start_point: Point,
        flagpole_end_point: Point,
        color: Tuple[int, int, int] = (0, 0, 0),
    ):
        """
        Draw a small flagpole between two points.

        :param image: Image
        :param flagpole_start_point: Start of the flagpole
        :param flagpole_end_point: End of the flagpole
        :return: Image
        """
        return cv2.line(
            image,
            flagpole_start_point.as_int_tuple(),
            flagpole_end_point.as_int_tuple(),
            color=color,
            thickness=2,
        )

    def newline(self):
        """Move the cursor to the next line."""
        self.cursor_pos.x = 0
        self.cursor_pos.y += self.line_height + self.content_margin

    def set_cursor_pos(self, cursor_pos: Optional[Point] = None) -> None:
        """
        Move the cursor to a new position.

        :param cursor_pos: New position of the cursor; (0,0) if not specified.
        """
        if cursor_pos is None:
            cursor_pos = Point(0, 0)

        self.cursor_pos = cursor_pos


class ShapeDrawer(DrawerEntity[AnnotationScene]):
    """
    ShapeDrawer to draw any shape on a numpy array. Will overlay the shapes in the same way that the UI does.

    :param show_count: Whether or not to render the amount of objects on screen in the top left.
    :param is_one_label: Whether there is only one label present in the project.
    """

    def __init__(
        self,
        show_count: bool,
        is_one_label: bool,
        show_labels: bool,
        show_confidence: bool,
    ):
        super().__init__()
        self.show_labels = show_labels
        self.show_confidence = show_confidence
        self.show_count = show_count
        self.is_one_label = is_one_label

        if self.is_one_label and not self.show_count:
            self.show_labels = False

        self.shape_drawers = [
            self.RectangleDrawer(self.show_labels, self.show_confidence),
            self.PolygonDrawer(self.show_labels, self.show_confidence),
            self.EllipseDrawer(self.show_labels, self.show_confidence),
        ]

        # Always show global labels, especially if shape labels are disabled (because of is_one_label).
        self.top_left_drawer = self.TopLeftDrawer(
            True, self.show_confidence, self.is_one_label
        )

    def draw(
        self,
        image: np.ndarray,
        scene: Union[AnnotationScene, Prediction],
        labels: List[ScoredLabel],
        fill_shapes: bool = True,
    ) -> np.ndarray:
        """
        Use a compatible drawer to draw all shapes of an annotation to the corresponding image.

        Also render a label in the top left if we need to.

        :param image: Numpy image, one frame of a video on which to draw
        :param scene: AnnotationScene scene corresponding to this particular frame of the video
        :param labels: Can be passed as an empty list since they are already present in annotation_scene
        :return: Modified image.
        """
        num_annotations = 0

        self.top_left_drawer.set_cursor_pos()

        for annotation in scene.annotations:
            if isinstance(annotation.shape, Rectangle) and annotation.shape.is_full_box(
                image.shape[1], image.shape[0]
            ):
                # If is_one_label is activated, don't draw the labels here
                # because we will draw them again outside the loop.
                if not self.is_one_label:
                    image = self.top_left_drawer.draw(image, annotation, labels=[])
            else:
                num_annotations += 1
                for drawer in self.shape_drawers:
                    if (
                        type(annotation.shape) in drawer.supported_types
                        and len(annotation.labels) > 0
                    ):
                        image = drawer.draw(
                            image,
                            annotation.shape,
                            labels=annotation.labels,
                            fill_shapes=fill_shapes,
                        )
        if self.is_one_label:
            image = self.top_left_drawer.draw_labels(image, scene.get_labels())
        if self.show_count:
            image = self.top_left_drawer.draw_annotation_count(image, num_annotations)
        return image

    class TopLeftDrawer(Helpers, DrawerEntity[Annotation]):
        """Draws labels in an image's top left corner."""

        def __init__(self, show_labels, show_confidence, is_one_label):
            super().__init__()
            self.show_labels = show_labels
            self.show_confidence = show_confidence
            self.is_one_label = is_one_label

        def draw(
            self, image: np.ndarray, annotation: Annotation, labels: List[ScoredLabel]
        ) -> np.ndarray:
            """
            Draw the labels of a shape in the image top left corner.

            :param image: Image
            :param annotation: Annotation
            :param labels: (Unused) labels to be drawn on the image
            :return: Image with label on top.
            """
            return self.draw_labels(image, annotation.labels)

        def draw_labels(
            self, image: np.ndarray, labels: Sequence[Union[Label, ScoredLabel]]
        ) -> np.ndarray:
            """
            Draw the labels in the image top left corner.

            :param image: Image
            :param labels: Sequence of labels
            :return: Image with label on top.
            """
            show_confidence = self.show_confidence if not self.is_one_label else False

            draw_command, _, _ = self.generate_draw_command_for_labels(
                labels, image, self.show_labels, show_confidence
            )

            image = draw_command(image)

            if len(labels) > 0:
                self.newline()

            return image

        def draw_annotation_count(
            self, image: np.ndarray, num_annotations: int
        ) -> np.ndarray:
            """
            Draw the number of annotations to the top left corner of the image.

            :param image: Image
            :param num_annotations: Number of annotations
            :return: Image with annotation count on top.
            """
            text = f"Count: {num_annotations}"
            color = self.yellow

            text_scale = self.generate_text_scale(image)
            draw_command, _, _ = self.generate_draw_command_for_text(
                text, text_scale, self.top_left_box_thickness, color
            )
            image = draw_command(image)

            self.newline()

            return image

    class RectangleDrawer(Helpers, DrawerEntity[Rectangle]):
        """Draws rectangles."""

        supported_types = [Rectangle]

        def __init__(self, show_labels, show_confidence):
            super().__init__()
            self.show_labels = show_labels
            self.show_confidence = show_confidence
            self.label_offset_box_shape = 0

        def draw(
            self,
            image: np.ndarray,
            entity: Rectangle,
            labels: List[ScoredLabel],
            fill_shapes: bool = True,
        ) -> np.ndarray:
            """
            Draw a rectangle on the image along with labels.

            :param image: Image to draw on.
            :param entity: Rectangle to draw.
            :param labels: List of labels.
            :param fill_shapes: Whether to fill the shapes with color.
            :return: Image with rectangle drawn on it.
            """
            base_color = labels[0].color_tuple

            # Draw the rectangle on the image
            x1, y1 = int(entity.x), int(entity.y)
            x2, y2 = int(entity.x + entity.width), int(entity.y + entity.height)
            if fill_shapes:
                image = self.draw_transparent_rectangle(
                    image, x1, y1, x2, y2, base_color, self.alpha_shape
                )
            image = cv2.rectangle(
                img=image, pt1=(x1, y1), pt2=(x2, y2), color=base_color, thickness=2
            )

            (
                draw_command,
                content_width,
                content_height,
            ) = self.generate_draw_command_for_labels(
                labels, image, self.show_labels, self.show_confidence
            )

            # Generate a command to draw the list of labels
            # and compute the actual size of the list of labels.
            y_coord = y1 - self.label_offset_box_shape - content_height
            x_coord = x1

            # put label inside if it is out of bounds at the top of the shape, and shift label to left if needed
            if y_coord < self.top_margin * image.shape[0]:
                y_coord = y1 + self.label_offset_box_shape
            if x_coord + content_width > image.shape[1]:
                x_coord = x2 - content_width

            # Draw the list of labels.
            self.set_cursor_pos(Point(x_coord, y_coord))
            image = draw_command(image)
            return image

    class EllipseDrawer(Helpers, DrawerEntity[Ellipse]):
        """Draws ellipses."""

        supported_types = [Ellipse]

        def __init__(self, show_labels, show_confidence):
            super().__init__()
            self.show_labels = show_labels
            self.show_confidence = show_confidence

        def draw(
            self,
            image: np.ndarray,
            entity: Ellipse,
            labels: List[ScoredLabel],
            fill_shapes: bool = True,
        ) -> np.ndarray:
            """
            Draw the ellipse on the image.

            :param image: Image to draw on.
            :param entity: Ellipse to draw.
            :param labels: Labels to draw.
            :param fill_shapes: Whether to fill the shapes with color.
            :return: Image with the ellipse drawn on it.
            """
            base_color = labels[0].color_tuple
            if entity.width > entity.height:
                axes = (
                    int(entity.major_axis * image.shape[1]),
                    int(entity.minor_axis * image.shape[0]),
                )
            else:
                axes = (
                    int(entity.major_axis * image.shape[0]),
                    int(entity.minor_axis * image.shape[1]),
                )
            center = (
                int(entity.x_center * image.shape[1]),
                int(entity.y_center * image.shape[0]),
            )
            # Draw the shape on the image
            alpha = self.alpha_shape
            if fill_shapes:
                overlay = cv2.ellipse(
                    img=image.copy(),
                    center=center,
                    axes=axes,
                    angle=0,
                    startAngle=0,
                    endAngle=360,
                    color=base_color,
                    thickness=cv2.FILLED,
                )
                result_without_border = cv2.addWeighted(
                    overlay, alpha, image, 1 - alpha, 0
                )
            else:
                result_without_border = image
            result_with_border = cv2.ellipse(
                img=result_without_border,
                center=center,
                axes=axes,
                angle=0,
                startAngle=0,
                endAngle=360,
                color=base_color,
                lineType=cv2.LINE_AA,
            )

            # Generate a command to draw the list of labels
            # and compute the actual size of the list of labels.
            (
                draw_command,
                content_width,
                content_height,
            ) = self.generate_draw_command_for_labels(
                labels, image, self.show_labels, self.show_confidence
            )

            # get top left corner of imaginary bbox around circle
            offset = self.label_offset_box_shape
            x_coord = entity.x1 * image.shape[1]
            y_coord = entity.y1 * image.shape[0] - offset - content_height

            flagpole_end_point = Point(entity.get_center_point())

            # put label at bottom if it is out of bounds at the top of the shape, and shift label to left if needed
            if y_coord < self.top_margin * image.shape[0]:
                y_coord = (
                    (entity.y1 * image.shape[0]) + (entity.y2 * image.shape[0]) + offset
                )
                flagpole_start_point = Point(x_coord + 1, y_coord)
            else:
                flagpole_start_point = Point(x_coord + 1, y_coord + content_height)

            if x_coord + content_width > result_with_border.shape[1]:
                # The list of labels is too close to the right side of the image.
                # Move it slightly to the left.
                x_coord = result_with_border.shape[1] - content_width

            # Draw the list of labels and a small flagpole.
            self.set_cursor_pos(Point(x_coord, y_coord))
            image = draw_command(result_with_border)
            image = self.draw_flagpole(
                image, flagpole_start_point, flagpole_end_point, labels[0].color_tuple
            )

            return image

    class PolygonDrawer(Helpers, DrawerEntity[Polygon]):
        """Draws polygons."""

        supported_types = [Polygon, RotatedRectangle]

        def __init__(self, show_labels, show_confidence):
            super().__init__()
            self.show_labels = show_labels
            self.show_confidence = show_confidence

        def draw(
            self,
            image: np.ndarray,
            entity: Union[Polygon, RotatedRectangle],
            labels: List[ScoredLabel],
            fill_shapes: bool = True,
        ) -> np.ndarray:
            """
            Draw polygon and labels on image.

            :param image: Image to draw on.
            :param entity: Polygon to draw.
            :param labels: List of labels to draw.
            :param fill_shapes: Whether to fill the shapes with color.
            :return: Image with polygon drawn on it.
            """
            if isinstance(entity, RotatedRectangle):
                entity = entity.to_polygon()
            base_color = labels[0].color_tuple

            # Draw the shape on the image
            alpha = self.alpha_shape
            contours = np.array(
                [[point.x, point.y] for point in entity.points],
                dtype=np.int32,
            )
            if fill_shapes:
                overlay = cv2.drawContours(
                    image=image.copy(),
                    contours=[contours],
                    contourIdx=-1,
                    color=base_color,
                    thickness=cv2.FILLED,
                )
                result_without_border = cv2.addWeighted(
                    overlay, alpha, image, 1 - alpha, 0
                )
            else:
                result_without_border = image
            result_with_border = cv2.drawContours(
                image=result_without_border,
                contours=[contours],
                contourIdx=-1,
                color=base_color,
                thickness=2,
                lineType=cv2.LINE_AA,
            )

            # Generate a command to draw the list of labels
            # and compute the actual size of the list of labels.
            (
                draw_command,
                content_width,
                content_height,
            ) = self.generate_draw_command_for_labels(
                labels, image, self.show_labels, self.show_confidence
            )

            # get point in the center of imaginary bbox around polygon
            x_coords, y_coords = zip(*[(point[0], point[1]) for point in contours])
            x_coord = (2 * np.median(x_coords) + np.min(x_coords)) / 3
            y_coord = min(y_coords) - self.label_offset_box_shape - content_height

            # end point = Y is the median poly Y, x offset to make line flush with text rectangle
            flagpole_end_point = Point(x_coord + 1, np.median(y_coords))

            if y_coord < self.top_margin * image.shape[0]:
                # The polygon is too close to the top of the image.
                # Draw the labels underneath the polygon instead.
                y_coord = max(y_coords) + self.label_offset_box_shape
                flagpole_start_point = Point(x_coord + 1, y_coord)
            else:
                flagpole_start_point = Point(x_coord + 1, y_coord + content_height)

            if x_coord + content_width > result_with_border.shape[1]:
                # The list of labels is too close to the right side of the image.
                # Move it slightly to the left.
                x_coord = result_with_border.shape[1] - content_width

            # Draw the list of labels and a small flagpole.
            self.set_cursor_pos(Point(x_coord, y_coord))
            image = draw_command(result_with_border)
            image = self.draw_flagpole(
                image, flagpole_start_point, flagpole_end_point, labels[0].color_tuple
            )

            return image



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import List, Optional

import attr
import numpy as np

from geti_sdk.data_models import Annotation, Label, Prediction

from .region_of_interest import ROI


@attr.define
class IntermediateInferenceResult:
    """
    Inference results for intermediate tasks in the pipeline
    """

    prediction: Prediction
    image: np.ndarray
    rois: Optional[List[ROI]] = None

    @property
    def image_width(self) -> int:
        """
        Return the width of the image to which the InferenceResult applies.

        :return: Integer representing the width of the image, in pixels
        """
        return self.image.shape[1]

    @property
    def image_height(self) -> int:
        """
        Return the height of the image to which the InferenceResult applies.

        :return: Integer representing the height of the image, in pixels
        """
        return self.image.shape[0]

    def filter_rois(self, label: Optional[Label] = None) -> List[ROI]:
        """
        Filter the ROIs for the inference results based on an input label.

        :param label: Label to retrieve the ROIs for. If left as None, all the ROIs
            belonging to the inference result are returned
        :return: List of ROIs containing an object with the specified label
        """
        if label is None:
            return self.rois
        return [roi for roi in self.rois if label.name in roi.label_names]

    def generate_views(self, rois: Optional[List[ROI]] = None) -> List[np.ndarray]:
        """
        Generate a list of image views holding the pixel data for the ROIs produced
        by the last local-label task in the pipeline.

        :param rois: Optional list of ROIs to return the views for. If left as None,
            views for all ROIs are returned.
        :return: List of numpy arrays containing the pixel data for the ROI's in the
            list of ROI's associated with this inference result
        """
        if self.rois is None:
            return [self.image]

        if rois is not None:
            rois_to_get = [roi.shape for roi in self.rois if roi in rois]
        else:
            rois_to_get = [roi.shape for roi in self.rois]

        views: List[np.ndarray] = []
        for roi in rois_to_get:
            y0, y1 = int(roi.y), int(roi.y + roi.height)
            x0, x1 = int(roi.x), int(roi.x + roi.width)
            if len(self.image.shape) == 3:
                views.append(self.image[y0:y1, x0:x1, :])
            elif len(self.image.shape) == 2:
                views.append(self.image[y0:y1, x0:x1])
            else:
                raise ValueError(
                    f"Unexpected image shape: {self.image.shape}. Unable to generate "
                    f"image views"
                )
        return views

    def append_annotation(self, annotation: Annotation, roi: ROI):
        """
        Append an Annotation instance to the prediction results, taking into account
        the ROI for which the annotation was predicted.

        This method can be used to add annotations produced by a downstream local task
        to the prediction results

        :param annotation: Annotation to append to the inference results
        :param roi: ROI in which the prediction was made
        """
        absolute_shape = annotation.shape.to_absolute_coordinates(parent_roi=roi.shape)
        self.prediction.append(
            Annotation(labels=annotation.labels, shape=absolute_shape)
        )

    def extend_annotations(self, annotations: List[Annotation], roi: ROI):
        """
        Extend the list of annotations for the current prediction results, taking
        into account the ROI for which the annotation was predicted.

        This method can be used to add labels produced by a global downstream task to
        the ROI output of it's upstream local task

        :param annotations: List of annotations holding the labels to append
        :param roi: ROI for which the annotations are predicted
        """
        for annotation in annotations:
            annotation.shape = roi.original_shape
        self.prediction.extend(annotations)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
from enum import Enum


class DeploymentState(Enum):
    """
    Enum representing the status of a deployment creation process for an Intel® Geti™
    project.
    """

    NONE = "NONE"
    PREPARING = "PREPARING"
    DONE = "DONE"
    FAILED = "FAILED"

    def __str__(self):
        """
        Return the string representation of the DeploymentState instance.
        """
        return self.value



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from pprint import pformat
from typing import Any, ClassVar, Dict, List, Optional, Union

import attr

from geti_sdk.data_models.enums.configuration_enums import (
    ParameterDataType,
    ParameterInputType,
)
from geti_sdk.data_models.utils import (
    attr_value_serializer,
    remove_null_fields,
    str_to_enum_converter,
)


@attr.define
class ConfigurableParameter:
    """
    Representation of a generic configurable parameter in GETi.

    :var data_type: Data type for the parameter. Can be integer, float, string or
        boolean
    :var default_value: Default value for the parameter
    :var description: Human readable description of the parameter
    :var editable: Boolean indicating whether this parameter is editable (True) or not
        (False)
    :var header: Human readable name for the parameter
    :var name: system name for the parameter
    :var template_type: Indicates whether the parameter takes free input (`input`)
        or the value has to be selected from a list of options (`selectable`)
    :var value: The current value for the parameter
    :var ui_rules: Dictionary representing rules for logic processing in the UI,
        based on parameter values
    :var warning: Optional warning message pointing out possible risks of changing the
        parameter
    """

    _identifier_fields: ClassVar[List[str]] = []
    _non_minimal_fields: ClassVar[List[str]] = [
        "default_value",
        "description",
        "editable",
        "header",
        "warning",
        "ui_rules",
    ]

    name: str
    value: Union[str, bool, float, int]
    data_type: Optional[str] = attr.field(
        default=None, converter=str_to_enum_converter(ParameterDataType)
    )
    default_value: Optional[Union[str, bool, float, int]] = None
    description: Optional[str] = None
    editable: Optional[bool] = None
    header: Optional[str] = None
    template_type: Optional[str] = attr.field(
        default=None, converter=str_to_enum_converter(ParameterInputType)
    )
    ui_rules: Optional[Dict[str, Any]] = None
    warning: Optional[str] = None
    auto_hpo_state: Optional[str] = None
    auto_hpo_value: Optional[Union[str, bool, float, int]] = None

    def to_dict(self) -> Dict[str, Any]:
        """
        Return the dictionary representation of the ConfigurableParameter object.
        """
        return attr.asdict(self, recurse=True, value_serializer=attr_value_serializer)

    @property
    def summary(self) -> str:
        """
        Return a string containing a very brief summary of the ConfigurableParameter
        object.

        :return: string holding a very short summary of the ConfigurableParameter
        """
        summary_str = f"{self.header} -- Name: {self.name} -- Value: {self.value}"
        return summary_str

    @property
    def overview(self) -> str:
        """
        Return a string that shows an overview of the configurable parameter. This
        still shows all the metadata of the parameter. If less details are required,
        please use the `summary` property

        :return: String holding an overview of the configurable parameter
        """
        overview_dict = self.to_dict()
        remove_null_fields(overview_dict)
        overview_dict.pop("ui_rules")
        return pformat(overview_dict)


@attr.define
class ConfigurableBoolean(ConfigurableParameter):
    """
    Representation of a configurable boolean in GETi.
    """

    default_value: Optional[bool] = attr.field(default=None, kw_only=True)
    value: bool = attr.field(kw_only=True)


@attr.define
class ConfigurableInteger(ConfigurableParameter):
    """
    Representation of a configurable integer in GETi.

    :var min_value: Minimum value allowed to be set for the configurable integer
    :var max_value: Maximum value allowed to be set for the configurable integer
    """

    default_value: Optional[int] = attr.field(default=None, kw_only=True)
    value: int = attr.field(kw_only=True)
    min_value: Optional[int] = attr.field(default=None, kw_only=True)
    max_value: Optional[int] = attr.field(default=None, kw_only=True)


@attr.define
class ConfigurableFloat(ConfigurableParameter):
    """
    Representation of a configurable float in GETi.

    :var min_value: Minimum value allowed to be set for the configurable float
    :var max_value: Maximum value allowed to be set for the configurable float
    """

    default_value: Optional[float] = attr.field(kw_only=True, default=None)
    value: float = attr.field(kw_only=True)
    min_value: float = attr.field(kw_only=True)
    max_value: float = attr.field(kw_only=True)
    step_size: Optional[float] = attr.field(
        kw_only=True, default=None
    )  # Added in Geti v1.6


@attr.define
class SelectableFloat(ConfigurableParameter):
    """
    Representation of a float selectable configurable parameter in GETi.

    :var options: List of options that the selectable float is allowed to take
    """

    default_value: Optional[float] = attr.field(kw_only=True, default=None)
    value: float = attr.field(kw_only=True)
    options: List[float] = attr.field(kw_only=True)


@attr.define
class SelectableString(ConfigurableParameter):
    """
    Representation of a string selectable configurable parameter in GETi.

    :var options: List of options that the selectable string is allowed to take
    """

    default_value: Optional[str] = attr.field(kw_only=True, default=None)
    enum_name: str = attr.field(kw_only=True)
    value: str = attr.field(kw_only=True)
    options: List[str] = attr.field(kw_only=True)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from pprint import pformat
from typing import Any, Dict, Optional

import attr

from geti_sdk.data_models.enums import TaskType
from geti_sdk.data_models.utils import str_to_optional_enum_converter

from .utils import attr_value_serializer, remove_null_fields


@attr.define
class Algorithm:
    """
    Representation of a supported algorithm on the Intel® Geti™ platform.
    """

    model_size: str
    model_template_id: str
    gigaflops: float
    name: Optional[str] = None
    summary: Optional[str] = None
    task_type: Optional[str] = attr.field(
        default=None, converter=str_to_optional_enum_converter(TaskType)
    )
    supports_auto_hpo: Optional[bool] = None
    default_algorithm: Optional[bool] = None  # Added in Geti v1.16
    performance_category: Optional[str] = None  # Added in Geti v1.9
    lifecycle_stage: Optional[str] = None  # Added in Geti v1.9

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the Algorithm to a dictionary representation.

        :return: Dictionary holding the algorithm data
        """
        output_dict = attr.asdict(self, value_serializer=attr_value_serializer)
        return output_dict

    @property
    def overview(self) -> str:
        """
        Return a string that shows an overview of the Algorithm properties.

        :return: String holding an overview of the algorithm
        """
        overview_dict = self.to_dict()
        remove_null_fields(overview_dict)
        return pformat(overview_dict)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
from typing import Any, Dict, List, cast

import attr
from omegaconf import OmegaConf

from geti_sdk.data_models import Annotation, AnnotationScene
from geti_sdk.data_models.enums import ShapeType
from geti_sdk.data_models.label import ScoredLabel
from geti_sdk.data_models.media import MediaType
from geti_sdk.data_models.media_identifiers import (
    ImageIdentifier,
    MediaIdentifier,
    VideoFrameIdentifier,
)
from geti_sdk.data_models.shapes import (
    Ellipse,
    Point,
    Polygon,
    Rectangle,
    RotatedRectangle,
    Shape,
)
from geti_sdk.data_models.task_annotation_state import TaskAnnotationState
from geti_sdk.data_models.utils import (
    attr_value_serializer,
    remove_null_fields,
    str_to_media_type,
    str_to_shape_type,
)

SHAPE_TYPE_MAPPING = {
    ShapeType.ELLIPSE: Ellipse,
    ShapeType.RECTANGLE: Rectangle,
    ShapeType.POLYGON: Polygon,
    ShapeType.ROTATED_RECTANGLE: RotatedRectangle,
}
MEDIA_IDENTIFIER_MAPPING = {
    MediaType.IMAGE: ImageIdentifier,
    MediaType.VIDEO_FRAME: VideoFrameIdentifier,
}


class AnnotationRESTConverter:
    """
    Class to convert REST representations of annotations into AnnotationScene entities.
    """

    @staticmethod
    def to_dict(
        annotation_scene: AnnotationScene, deidentify: bool = True
    ) -> Dict[str, Any]:
        """
        Convert an AnnotationScene to a dictionary. By default, removes any ID
        fields in the output dictionary

        :param annotation_scene: AnnotationScene object to convert
        :param deidentify: True to remove any unique database ID fields in the output,
            False to keep these fields. Defaults to True
        :return: Dictionary holding the serialized AnnotationScene data
        """
        if deidentify:
            annotation_scene.deidentify()
        annotation_dict = attr.asdict(
            annotation_scene, recurse=True, value_serializer=attr_value_serializer
        )
        remove_null_fields(annotation_dict)
        return annotation_dict

    @staticmethod
    def _shape_from_dict(input_dict: Dict[str, Any]) -> Shape:
        """
        Convert a dictionary representing a shape to a Shape object.

        :param input_dict:
        :return: Shape corresponding to the input dict
        """
        input_copy = copy.deepcopy(input_dict)
        type_ = str_to_shape_type(input_copy.get("type"))
        class_type = SHAPE_TYPE_MAPPING[type_]
        if issubclass(class_type, Polygon):
            points_dicts = input_copy.pop("points")
            points = [Point(**point) for point in points_dicts]
            input_copy.update({"points": points})
        return class_type(**input_copy)

    @staticmethod
    def _scored_label_from_dict(input_dict: Dict[str, Any]) -> ScoredLabel:
        """
        Create a ScoredLabel object from an input dictionary.

        :param input_dict:
        :return:
        """
        label_dict_config = OmegaConf.create(input_dict)
        schema = OmegaConf.structured(ScoredLabel)
        values = OmegaConf.merge(schema, label_dict_config)
        return cast(ScoredLabel, OmegaConf.to_object(values))

    @staticmethod
    def annotation_from_dict(input_dict: Dict[str, Any]) -> Annotation:
        """
        Convert a dictionary representing an annotation to an Annotation object.

        :param input_dict:
        :return:
        """
        input_copy = copy.deepcopy(input_dict)
        labels: List[ScoredLabel] = []
        for label in input_dict["labels"]:
            labels.append(AnnotationRESTConverter._scored_label_from_dict(label))
        shape = AnnotationRESTConverter._shape_from_dict(input_dict["shape"])
        input_copy.update({"labels": labels, "shape": shape})
        return Annotation(**input_copy)

    @staticmethod
    def _media_identifier_from_dict(input_dict: Dict[str, Any]) -> MediaIdentifier:
        """
        Convert a dictionary representing a media identifier to a MediaIdentifier
        object.

        :param input_dict:
        :return:
        """
        if isinstance(input_dict, MediaIdentifier):
            return input_dict
        type_ = str_to_media_type(input_dict["type"])
        identifier_type = MEDIA_IDENTIFIER_MAPPING[type_]
        return identifier_type(**input_dict)

    @staticmethod
    def from_dict(annotation_scene: Dict[str, Any]) -> AnnotationScene:
        """
        Create an AnnotationScene object from a dictionary returned by the
        /annotations REST endpoint in the Intel® Geti™ platform.

        :param annotation_scene: dictionary representing an AnnotationScene, which
            contains all annotations for a certain media entity
        :return: AnnotationScene object
        """
        input_copy = copy.deepcopy(annotation_scene)
        annotations: List[Annotation] = []
        for annotation in annotation_scene["annotations"]:
            if not isinstance(annotation, Annotation):
                annotations.append(
                    AnnotationRESTConverter.annotation_from_dict(annotation)
                )
            else:
                annotations.append(annotation)
        media_identifier = AnnotationRESTConverter._media_identifier_from_dict(
            annotation_scene["media_identifier"]
        )
        input_copy.update(
            {"annotations": annotations, "media_identifier": media_identifier}
        )
        if "annotation_state_per_task" in annotation_scene:
            annotation_states: List[TaskAnnotationState] = []
            for annotation_state in annotation_scene["annotation_state_per_task"]:
                annotation_states.append(TaskAnnotationState(**annotation_state))
            input_copy.update({"annotation_state_per_task": annotation_states})
        return AnnotationScene(**input_copy)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Any, Dict

from geti_sdk.data_models.model import Model, OptimizedModel
from geti_sdk.data_models.model_group import ModelGroup
from geti_sdk.utils import deserialize_dictionary


class ModelRESTConverter:
    """
    Class that handles conversion of Intel® Geti™ REST output for media entities to
    objects and vice versa.
    """

    @staticmethod
    def model_group_from_dict(input_dict: Dict[str, Any]) -> ModelGroup:
        """
        Convert a dictionary representing a model group to a ModelGroup object.

        :param input_dict: Dictionary representing a model group, as returned by the
            Intel® Geti™ /model_groups REST endpoint
        :return: ModelGroup object corresponding to the data in `input_dict`
        """
        return deserialize_dictionary(input_dict, output_type=ModelGroup)

    @staticmethod
    def model_from_dict(input_dict: Dict[str, Any]) -> Model:
        """
        Convert a dictionary representing a model to a Model object.

        :param input_dict: Dictionary representing a model, as returned by the
            Intel® Geti™ /model_groups/models REST endpoint
        :return: Model object corresponding to the data in `input_dict`
        """
        return deserialize_dictionary(input_dict, output_type=Model)

    @staticmethod
    def optimized_model_from_dict(input_dict: Dict[str, Any]) -> OptimizedModel:
        """
        Convert a dictionary representing an optimized model to a OptimizedModel object.

        :param input_dict: Dictionary representing an optimized model, as returned by
            the Intel® Geti™ /model_groups/models REST endpoint
        :return: OptimizedModel object corresponding to the data in `input_dict`
        """
        return deserialize_dictionary(input_dict, output_type=OptimizedModel)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
from datetime import datetime
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Type, TypeVar, Union

import attr
import cv2
import numpy as np

from geti_sdk.data_models.enums import AnnotationKind, MediaType, ShapeType, TaskType

EnumType = TypeVar("EnumType", bound=Enum)


def deidentify(instance: Any):
    """
    Set all identifier fields of an instance of an attr.define decorated class within the
    GETi REST DataModels to None.

    :param instance: Object to deidentify
    """
    for field in attr.fields(type(instance)):
        if field.name in instance._identifier_fields:
            setattr(instance, field.name, None)


def str_to_enum_converter(
    enum: Type[EnumType],
) -> Callable[[Union[str, EnumType]], EnumType]:
    """
    Construct a converter function to convert an input value into an instance of the
    Enum subclass passed in `enum`.

    :param enum: type of the Enum to which the converter should convert
    :return: Converter function that takes an input value and attempts to convert it
        into an instance of `enum`
    """

    def _converter(input_value: Union[str, EnumType]) -> EnumType:
        """
        Convert an input value to an instance of an Enum.

        :param input_value: Value to convert
        :return: Instance of the Enum
        """
        if isinstance(input_value, str):
            return enum(input_value)
        elif isinstance(input_value, enum):
            return input_value
        else:
            raise ValueError(
                f"Invalid argument! Cannot convert value {input_value} to Enum "
                f"{enum.__name__}"
            )

    return _converter


def str_to_optional_enum_converter(
    enum: Type[EnumType],
) -> Callable[[Union[str, EnumType]], EnumType]:
    """
    Construct a converter function to convert an input value into an instance of the
    Enum subclass passed in `enum`.

    :param enum: type of the Enum to which the converter should convert
    :return: Converter function that takes an input value and attempts to convert it
        into an instance of `enum`
    """

    def _converter(input_value: Optional[Union[str, EnumType]]) -> Optional[EnumType]:
        """
        Convert an input value to an instance of an Enum.

        :param input_value: Value to convert
        :return: Instance of the Enum
        """
        if isinstance(input_value, str):
            return enum(input_value)
        elif isinstance(input_value, enum):
            return input_value
        elif input_value is None:
            return None
        else:
            raise ValueError(
                f"Invalid argument! Cannot convert value {input_value} to Enum "
                f"{enum.__name__}"
            )

    return _converter


def str_to_enum_converter_by_name_or_value(
    enum: Type[EnumType], allow_none: bool = False
) -> Callable[[Union[str, EnumType]], EnumType]:
    """
    Construct a converter function to convert an input value into an instance of the
    Enum subclass passed in `enum`.

    This method attempts to convert both from the Enum value as well as it's name

    :param enum: type of the Enum to which the converter should convert
    :param allow_none: True to allow `None` as input value, i.e. for optional parameters
    :return: Converter function that takes an input value and attempts to convert it
        into an instance of `enum`
    """

    def _converter(input_value: Union[str, EnumType]) -> EnumType:
        """
        Convert an input value to an instance of an Enum.

        :param input_value: Value to convert
        :return: Instance of the Enum
        """
        if isinstance(input_value, str):
            try:
                return enum(input_value)
            except ValueError:
                return enum[input_value]
        elif isinstance(input_value, enum):
            return input_value
        if (input_value is None) and allow_none:
            return None
        else:
            raise ValueError(
                f"Invalid argument! Cannot convert value {input_value} to Enum "
                f"{enum.__name__}"
            )

    return _converter


def str_to_task_type(task_type: Union[str, TaskType]) -> TaskType:
    """
    Convert an input string to a task type.

    :param task_type:
    :return: TaskType instance corresponding to `task_type`
    """
    if isinstance(task_type, str):
        return TaskType(task_type)
    else:
        return task_type


def str_to_media_type(media_type: Union[str, MediaType]) -> MediaType:
    """
    Convert an input string to a media type.

    :param media_type:
    :return: MediaType instance corresponding to `media_type`
    """
    if isinstance(media_type, str):
        return MediaType(media_type)
    else:
        return media_type


def str_to_shape_type(shape_type: Union[str, ShapeType]) -> ShapeType:
    """
    Convert an input string to a shape type.

    :param shape_type:
    :return: ShapeType instance corresponding to `shape_type`
    """
    if isinstance(shape_type, str):
        return ShapeType(shape_type)
    else:
        return shape_type


def str_to_annotation_kind(
    annotation_kind: Union[str, AnnotationKind]
) -> AnnotationKind:
    """
    Convert an input string to an annotation kind.

    :param annotation_kind:
    :return: AnnotationKind instance corresponding to `annotation_kind`
    """
    if isinstance(annotation_kind, str):
        return AnnotationKind(annotation_kind)
    else:
        return annotation_kind


def str_to_datetime(datetime_str: Optional[Union[str, datetime]]) -> Optional[datetime]:
    """
    Convert a string to a datetime.

    :param datetime_str: string containing the datetime, in isoformat.
    :return: datetime instance
    """
    if isinstance(datetime_str, str):
        try:
            return datetime.fromisoformat(datetime_str)
        except ValueError:
            logging.debug(
                f"Unable to convert str '{datetime_str}' to datetime, converter "
                f"returns None instead."
            )
            return None
    elif isinstance(datetime_str, datetime):
        return datetime_str
    elif datetime_str is None:
        return None


def attr_value_serializer(instance, field, value):
    """
    Convert a value in an attr.define decorated class to string representation, used
    while converting the attrs object to a dictionary.

    Converts Enums and datetime objects to string representation

    :param instance:
    :param field:
    :param value:
    :return:
    """
    if isinstance(value, Enum):
        return str(value)
    elif isinstance(value, datetime):
        return datetime.isoformat(value)
    else:
        return value


def numpy_from_buffer(buffer: bytes) -> np.ndarray:
    """
    Convert a bytes string representing an image into a numpy array.

    :param buffer: Bytes object to convert
    :return: Numpy.ndarray containing the numpy data from the image
    """
    numpy_array = np.frombuffer(buffer, dtype=np.uint8)
    return cv2.imdecode(numpy_array, cv2.IMREAD_COLOR)


def round_dictionary(
    input_data: Union[Dict[str, Any], List[Any]], decimal_places: int = 3
) -> Union[Dict[str, Any], List[Any]]:
    """
    Convert all floats in a dictionary to string representation, rounded to
    `decimal_places`.

    :param input_data: Input dictionary to convert and round
    :param decimal_places: Number of decimal places to round to. Defaults to 3
    :return: dictionary with all floats rounded
    """
    if isinstance(input_data, dict):
        for key, value in input_data.items():
            if isinstance(value, float):
                input_data[key] = f"{value:.{decimal_places}f}"
            elif isinstance(value, (dict, list)):
                input_data[key] = round_dictionary(value, decimal_places=decimal_places)
        return input_data
    elif isinstance(input_data, list):
        new_list = []
        for item in input_data:
            if isinstance(item, float):
                new_list.append(f"{item:.{decimal_places}f}")
            elif isinstance(item, (dict, list)):
                new_list.append(round_dictionary(item, decimal_places=decimal_places))
        return new_list


def round_to_n_digits(n: int) -> Callable[[float], float]:
    """
    Return a function to round an input number to n digits.

    :param n: Number of digits to round to
    :return: Callable that, when called with a number as input, will round the number
        to n digits.
    """

    def _n_digit_rounder(value: float) -> float:
        return round(value, ndigits=n)

    return _n_digit_rounder


def remove_null_fields(input: Any):
    """
    Remove fields that have 'None' or an emtpy string '' as their value from a
    dictionary.

    NOTE: This function modifies the input dictionary in place

    :param input: Dictionary to remove the null fields from
    """
    if isinstance(input, dict):
        for key, value in list(input.items()):
            if isinstance(value, dict):
                remove_null_fields(value)
            elif value is None or value == "":
                input.pop(key)
            elif isinstance(value, list):
                for item in value:
                    remove_null_fields(item)
    elif isinstance(input, list):
        for item in input:
            remove_null_fields(item)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
from pprint import pformat
from typing import Any, ClassVar, Dict, List, Optional

import attr

from geti_sdk.data_models.enums.task_type import (
    ANOMALY_TASK_TYPES,
    GLOBAL_TASK_TYPES,
    NON_TRAINABLE_TASK_TYPES,
    TaskType,
)
from geti_sdk.data_models.label import Label
from geti_sdk.data_models.utils import (
    attr_value_serializer,
    deidentify,
    remove_null_fields,
    str_to_task_type,
)


@attr.define
class Task:
    """
    Representation of a Task in GETi.

    :var title: Title of the task
    :var task_type: Type of the task
    :var labels: List of labels belonging to the task
    :var id: unique database ID of the task
    :var label_schema_id: unique database ID of the label schema for the task
    """

    _identifier_fields: ClassVar[List[str]] = ["id", "label_schema_id"]

    title: str
    task_type: str = attr.field(converter=str_to_task_type)
    labels: Optional[List[Label]] = None
    label_schema_id: Optional[str] = None
    id: Optional[str] = None

    @property
    def type(self) -> TaskType:
        """
        Return the type of the task.

        This property is here to make sure that the type of task_type is derived
        correctly. The attribute self.task_type can be instantiated both as a string
        and as a TaskType

        :return: type of the task
        """
        task_type = self.task_type
        if not isinstance(task_type, TaskType):
            task_type = TaskType(task_type)
        return task_type

    @property
    def is_trainable(self) -> bool:
        """
        Return True if this Task represents a trainable task.

        :return: True if the task is trainable, False otherwise
        """
        return self.type not in NON_TRAINABLE_TASK_TYPES

    @property
    def is_global(self) -> bool:
        """
        Return True if this Task represents a trainable task that produces global
        labels, False otherwise.

        :return: True if the task produces global labels, False otherwise
        """
        return self.type in GLOBAL_TASK_TYPES

    @property
    def is_anomaly(self) -> bool:
        """
        Return True if this task is an anomaly task.

        :return: True if task is an anomaly type task, False otherwise
        """
        return self.type in ANOMALY_TASK_TYPES

    def deidentify(self) -> None:
        """
        Set all identifier fields for the task to None.
        """
        deidentify(self)
        if self.is_trainable:
            for label in self.labels:
                deidentify(label)

    def get_label_names(self, include_empty: bool = True) -> List[str]:
        """
        Return a list of label names for the task.

        :param include_empty: True to include the empty label (if present), False to
            exclude it
        :return: List of label names for the task
        """
        if include_empty:
            labels = [label.name for label in self.labels]
        else:
            labels = [label.name for label in self.labels if not label.is_empty]
        return labels

    def to_dict(self) -> Dict[str, Any]:
        """
        Return the dictionary representation of the task.

        :return:
        """
        return attr.asdict(self, recurse=True, value_serializer=attr_value_serializer)

    @property
    def overview(self) -> str:
        """
        Return a string that shows an overview of the task. This still shows all
        the detailed information of the task. If less details are required, please
        use the `summary` property

        :return: String holding an overview of the project
        """
        deidentified = copy.deepcopy(self)
        deidentified.deidentify()
        overview_dict = deidentified.to_dict()
        remove_null_fields(overview_dict)
        return pformat(overview_dict)

    @property
    def summary(self) -> str:
        """
        Return a string that gives a very brief summary of the task. This is the
        least detailed representation of the task, if more details are required
        please use the `overview` property

        :return: String holding a brief summary of the task
        """
        summary_str = f"Task: {self.title}\n  Type: {self.type} \n  Labels:\n"
        for label in self.labels:
            summary_str += (
                f"    Name: {label.name},  Group: {label.group},  "
                f"Parent: {label.parent_id}\n"
            )
        return summary_str

    def prepare_for_post(self) -> None:
        """
        Set all fields to None that are not valid for making a POST request to the
        /projects endpoint.

        :return:
        """
        if self.labels is not None:
            labels_indices_to_pop: List[int] = []
            for ii, label in enumerate(self.labels):
                if label.is_empty:
                    labels_indices_to_pop.append(ii)
                label.prepare_for_post()
            for index in labels_indices_to_pop:
                # Empty labels are not allowed to be specified explicitly in a POST
                # request
                self.labels.pop(index)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from __future__ import annotations

import os
from collections import UserList
from typing import Any, Dict, Generic, List, Type, TypeVar

from geti_sdk.data_models.media import Image, MediaItem, Video, VideoFrame
from geti_sdk.utils.serialization_helpers import deserialize_dictionary

MediaTypeVar = TypeVar("MediaTypeVar", Image, Video, VideoFrame)


class MediaList(UserList, Generic[MediaTypeVar]):
    """
    A list containing Intel® Geti™ media entities
    """

    @property
    def ids(self) -> List[str]:
        """
        Return a list of unique database IDs for all media items in the media list.
        """
        return [item.id for item in self.data]

    @property
    def names(self) -> List[str]:
        """
        Return a list of filenames for all media items in the media list.
        """
        return [item.name for item in self.data]

    def get_by_id(self, id_value: str) -> MediaItem:
        """
        Return the item with id `id_value` from the media list.
        """
        for item in self.data:
            if item.id == id_value:
                return item
        raise ValueError(f"Media list {self} does not contain item with ID {id_value}.")

    def get_by_filename(self, filename: str) -> MediaItem:
        """
        Return the item with name `filename` from the media list.
        """
        for item in self.data:
            if item.name == filename:
                return item
        raise ValueError(
            f"Media list {self} does not contain item with filename {filename}."
        )

    @property
    def media_type(self) -> Type[MediaTypeVar]:
        """
        Return the type of the media contained in this list.
        """
        if self.data:
            return type(self.data[0])
        else:
            raise ValueError("Cannot deduce media type from empty MediaList")

    @staticmethod
    def from_rest_list(
        rest_input: List[Dict[str, Any]], media_type: Type[MediaTypeVar]
    ) -> MediaList[MediaTypeVar]:
        """
        Create a MediaList instance from a list of media entities obtained from the
        Intel® Geti™ /media endpoints.

        :param rest_input: List of dictionaries representing media entities in Intel® Geti™
        :param media_type: Image or Video, type of the media entities that are to be
            converted.

        :return: MediaList holding the media entities contained in `rest_input`,
            where each entity is of type `media_type`
        """
        return MediaList[MediaTypeVar](
            [
                deserialize_dictionary(media_dict, output_type=media_type)
                for media_dict in rest_input
            ]
        )

    @property
    def has_duplicate_filenames(self) -> bool:
        """
        Return True if the media list contains at least two items that have the same
        filename, False otherwise.
        """
        filenames = [os.path.basename(name) for name in self.names]
        return len(set(filenames)) != len(filenames)



# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import os
from typing import List, Optional

from geti_sdk.data_models import Dataset, Project
from geti_sdk.http_session import GetiSession
from geti_sdk.utils import deserialize_dictionary


class DatasetClient:
    """
    Class to manage datasets for a certain Intel® Geti™ project.
    """

    def __init__(self, workspace_id: str, project: Project, session: GetiSession):
        self.session = session
        self.project = project
        self.workspace_id = workspace_id
        self.base_url = f"workspaces/{workspace_id}/projects/{project.id}/datasets"

    def create_dataset(self, name: str) -> Dataset:
        """
        Create a new dataset named `name` inside the project.

        :param name: Name of the dataset to create
        :return: The created dataset
        """
        request_data = {"name": name}
        response = self.session.get_rest_response(
            url=self.base_url,
            method="POST",
            data=request_data,
        )
        dataset = deserialize_dictionary(response, output_type=Dataset)
        self.project.datasets.append(dataset)
        return dataset

    def get_all_datasets(self) -> List[Dataset]:
        """
        Query the Intel® Geti™ server to retrieve an up to date list of datasets in
        the project.

        :return: List of current datasets in the project
        """
        if self.project.id is None:
            raise ValueError(
                "Project ID is not defined, please make sure that the project you pass "
                "contains a valid ID"
            )
        response = self.session.get_rest_response(url=self.base_url, method="GET")
        datasets = [
            deserialize_dictionary(dataset_dict, Dataset)
            for dataset_dict in response["datasets"]
        ]
        self.project.datasets = datasets
        return datasets

    def get_dataset_statistics(self, dataset: Dataset) -> dict:
        """
        Retrieve the media and annotation statistics for a particular dataset
        """
        response = self.session.get_rest_response(
            url=f"{self.base_url}/{dataset.id}/statistics", method="GET"
        )
        return response

    def get_dataset_by_name(self, dataset_name: str) -> Dataset:
        """
        Retrieve a dataset by name

        :param dataset_name: Name of the dataset to retrieve
        :return: Dataset object
        """
        dataset: Optional[Dataset] = None
        for ds in self.get_all_datasets():
            if ds.name == dataset_name:
                dataset = ds
        if dataset is None:
            raise ValueError(
                f"Dataset named '{dataset_name}' was not found in project "
                f"'{self.project.name}'"
            )
        return dataset

    def has_dataset_subfolders(self, path_to_folder: str) -> bool:
        """
        Check if a project folder has it's media folders organized according to the
        datasets in the project

        :param path_to_folder: Path to the project folder to check
        :return: True if the media folders in the project folder tree are organized
            according to the datasets in the project.
        """
        media_folder_names = ["images", "videos"]
        result: bool = True
        for folder_name in media_folder_names:
            full_path = os.path.join(path_to_folder, folder_name)
            if not os.path.exists(full_path):
                continue
            result *= self._media_folder_has_dataset_subfolders(full_path)
        return result

    def _media_folder_has_dataset_subfolders(self, path_to_folder: str) -> bool:
        """
        Check if a folder with media has subfolders organized according to the
        projects' datasets

        :param path_to_folder: Path to the media folder to check for dataset subfolders
        :return: True if the folder contains subfolders named according to the
            datasets in the project, False otherwise
        """
        if not os.path.isdir(path_to_folder):
            return False
        content = os.listdir(path_to_folder)
        for dataset in self.project.datasets:
            if dataset.name not in content:
                return False
            if not os.path.isdir(os.path.join(path_to_folder, dataset.name)):
                return False
        return True



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
Introduction
------------

The `utils` package contains utility functions used throughout the SDK.

Module contents
---------------
"""

from .algorithm_helpers import get_default_algorithm_info, get_supported_algorithms
from .credentials_helpers import get_server_details_from_env
from .dictionary_helpers import get_dict_key_from_value
from .label_helpers import (
    generate_classification_labels,
    generate_segmentation_labels,
    generate_unique_label_color,
)
from .plot_helpers import (
    show_image_with_annotation_scene,
    show_video_frames_with_annotation_scenes,
)
from .project_helpers import get_project_folder_name, get_task_types_by_project_type
from .serialization_helpers import deserialize_dictionary
from .workspace_helpers import get_default_workspace_id

__all__ = [
    "get_default_workspace_id",
    "generate_classification_labels",
    "generate_segmentation_labels",
    "get_dict_key_from_value",
    "get_task_types_by_project_type",
    "show_image_with_annotation_scene",
    "show_video_frames_with_annotation_scenes",
    "get_supported_algorithms",
    "deserialize_dictionary",
    "get_server_details_from_env",
    "get_project_folder_name",
    "generate_unique_label_color",
    "get_default_algorithm_info",
]



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from packaging.version import Version
from semver import Version as SemVersion


class GetiVersion:
    """
    Version identifier of the Intel Geti platform
    """

    _GETI10_TIMETAG = "20220910154208"
    _SC11_TIMETAG = "20220624125113"
    _SCMVP_TIMETAG = "20220129184214"

    def __init__(self, version_string: str) -> None:
        """
        Initialize the version

        :param version_string: String representation of the version, as returned by
            the Geti platform
        """
        version_parts = version_string.split("-")

        sem_version = SemVersion.parse(version_string)

        time_tag = version_parts[-1]
        base_version = Version(
            f"{sem_version.major}.{sem_version.minor}.{sem_version.patch}"
        )
        build_tag = sem_version.prerelease.replace(f"-{time_tag}", "")

        self.version = base_version
        self.build_tag = build_tag
        self.time_tag = time_tag

    def __gt__(self, other):
        """
        Return True if this GetiVersion instance is a later version than `other`

        :param other: GetiVersion object to compare with
        :raises: TypeError if `other` is not a GetiVersion instance
        :return: True if this instance corresponds to a later version of the Intel
            Geti platform than `other`
        """
        if not isinstance(other, GetiVersion):
            raise TypeError(
                f"Unsupported comparison operation, {other} is not a GetiVersion."
            )
        if self.is_geti and other.is_geti:
            if self.version != other.version:
                return self.version > other.version
            else:
                return self.time_tag > other.time_tag
        else:
            if self.is_geti and not other.is_geti:
                return True
            elif not self.is_geti and other.is_geti:
                return False
            else:
                return self.time_tag > other.time_tag

    def __lt__(self, other):
        """
        Return True if this GetiVersion instance is an earlier version than `other`

        :param other: GetiVersion object to compare with
        :raises: TypeError if `other` is not a GetiVersion instance
        :return: True if this instance corresponds to an earlier version of the Intel
            Geti platform than `other`
        """
        return not self >= other

    def __ge__(self, other):
        """
        Return True if this GetiVersion instance is a later or equivalent version than
        `other`

        :param other: GetiVersion object to compare with
        :raises: TypeError if `other` is not a GetiVersion instance
        :return: True if this instance corresponds to a later or equivalent version of
            the Intel Geti platform than `other`
        """
        return (self > other) or (self == other)

    def __le__(self, other):
        """
        Return True if this GetiVersion instance is an earlier or equivalent version
        than `other`

        :param other: GetiVersion object to compare with
        :raises: TypeError if `other` is not a GetiVersion instance
        :return: True if this instance corresponds to an earlier or equivalent version
            of the Intel Geti platform than `other`
        """
        return (self < other) or (self == other)

    def __eq__(self, other):
        """
        Return True if this GetiVersion instance is equal to `other`

        :param other: GetiVersion object to compare with
        :raises: TypeError if `other` is not a GetiVersion instance
        :return: True if this instance is equal to the GetiVersion passed in `other`
        """
        if not isinstance(other, GetiVersion):
            raise TypeError(
                f"Unsupported comparison operation, {other} is not a GetiVersion."
            )
        return self.version == other.version

    def __str__(self) -> str:
        """
        Return the Intel Geti version as a string

        :return: String containing the version
        """
        return f"{self.version}-{self.build_tag}-{self.time_tag}"

    def __repr__(self) -> str:
        """
        Return the string representation of the Intel Geti version

        :return: String representing the version
        """
        return f"GetiVersion({self.version}-{self.build_tag}-{self.time_tag})"

    @property
    def is_sc_mvp(self) -> bool:
        """
        Return True if the version corresponds to a platform on the SC MVP version of
        the software.
        """
        return (
            self.version == Version("1.0.0")
            and self._SCMVP_TIMETAG <= self.time_tag <= self._SC11_TIMETAG
        )

    @property
    def is_sc_1_1(self) -> bool:
        """
        Return True if the version corresponds to a platform on the SC v1.1 version of
        the software.
        """
        return (
            self.version == Version("1.1.0")
            and self._SC11_TIMETAG <= self.time_tag <= self._GETI10_TIMETAG
        )

    @property
    def is_geti(self) -> bool:
        """
        Return True if the version corresponds to any version of the Geti platform.
        Return False if it corresponds to any SC version.
        """
        return (
            self.version > Version("1.0.0b0")
            and self.time_tag >= self._GETI10_TIMETAG
            and not (self.is_sc_1_1 or self.is_sc_mvp)
        )


SC_MVP_VERSION = GetiVersion("1.0.0-release-20220129184214")
SC_11_VERSION = GetiVersion("1.1.0-release-20220624125113")
GETI_10_VERSION = GetiVersion("1.0.0-release-20221005164936")
GETI_11_VERSION = GetiVersion("1.1.0-release-20221125121144")
GETI_12_VERSION = GetiVersion("1.2.0-release-20230101120000")
GETI_15_VERSION = GetiVersion("1.5.0-release-20230504111017")
GETI_18_VERSION = GetiVersion("1.8.0-release-20231018022911")
GETI_114_VERSION = GetiVersion("1.14.0-release-20240131095302")
GETI_116_VERSION = GetiVersion("1.16.0-release-20240320101320")



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import List, Optional

import attr


@attr.define()
class Score:
    """
    Score attribute in TaskPerformance.

    :var value: Value of the score
    :var metric_type: The type of matrix the score represents
    """

    value: Optional[float] = None
    metric_type: Optional[str] = None


@attr.define()
class TaskPerformance:
    """
    Task Performance metrics in Intel® Geti™.

    :var task_id: Unique ID of the task to which this Performance metric
        applies.
    :var score: Score of the project or model for each task
    :var local_score: Accuracy of the model or project with respect to object
        localization for each task
    :var global_score: Accuracy of the model or project with respect to global
        classification of the full image for each task
    """

    task_id: Optional[str] = None
    score: Optional[Score] = None
    local_score: Optional[Score] = None
    global_score: Optional[Score] = None


@attr.define()
class Performance:
    """
    Performance metrics for a project or model in Intel® Geti™.

    :var score: Overall score of the project or model
    :var local_score: Accuracy of the model or project with respect to object
        localization
    :var global_score: Accuracy of the model or project with respect to global
        classification of the full image
    """

    score: Optional[float] = None
    local_score: Optional[float] = None
    global_score: Optional[float] = None
    task_performances: Optional[List[TaskPerformance]] = None



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import List, Optional

from geti_sdk.data_models.label import Label
from geti_sdk.data_models.label_group import LabelGroup, LabelGroupType


class LabelSchema:
    """
    The `LabelSchema` class defines the structure and properties of labels and label groups.

    :param label_groups: Optional list of `LabelGroup` objects representing the label groups in the schema
    """

    def __init__(self, label_groups: Optional[List[LabelGroup]] = None) -> None:
        """
        Initialize a new instance of the `LabelSchema` class.

        :param label_groups: Optional list of `LabelGroup` objects representing the label groups in the schema
        """
        self._groups = label_groups

    def get_labels(self, include_empty: bool = False) -> List[Label]:
        """
        Get the labels in the label schema.

        :param include_empty: Flag determining whether to include empty labels
        :return: List of all labels in the label schema
        """
        labels = {
            label
            for group in self._groups
            for label in group.labels
            if include_empty or not label.is_empty
        }
        return sorted(list(labels), key=lambda label: label.id)

    def get_groups(self, include_empty: bool = False) -> List[LabelGroup]:
        """
        Get the label groups in the label schema.

        :param include_empty: Flag determining whether to include empty label groups
        :return: List of all label groups in the label schema
        """
        if include_empty:
            return self._groups

        return [
            group
            for group in self._groups
            if group.group_type != LabelGroupType.EMPTY_LABEL
        ]



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
Introduction
------------

The `rest_clients` package contains clients for interacting with the various entities (
such as :py:class:`~geti_sdk.data_models.project.Project`,
:py:class:`~geti_sdk.data_models.media.Image` and
:py:class:`~geti_sdk.data_models.model.Model`)
on the Intel® Geti™ server.

All rest clients are initialized with a
:py:class:`~geti_sdk.http_session.geti_session.GetiSession` and a workspace id. The
:py:class:`~geti_sdk.rest_clients.project_client.project_client.ProjectClient`
can be initialized with just that, while all other clients are initialized
*per project* and thus take an additional `project` argument.

For example, to initialize the
:py:class:`~geti_sdk.rest_clients.media_client.image_client.ImageClient` for
a specific project and get a list of all images in the project, the following code
snippet can be used:

.. code-block:: python

   from geti_sdk import Geti
   from geti_sdk.rest_clients import ProjectClient, ImageClient

   geti = Geti(
     host="https://0.0.0.0", username="dummy_user", password="dummy_password"
   )

   project_client = ProjectClient(
       session=geti.session, workspace_id=geti.workspace_id
   )
   project = project_client.get_project_by_name(project_name='dummy_project')

   image_client = ImageClient(
       session=geti.session, workspace_id=geti.workspace_id, project=project
   )
   image_client.get_all_images()

Module contents
---------------

.. autoclass:: geti_sdk.rest_clients.project_client.project_client.ProjectClient
   :members:
   :undoc-members:

.. autoclass:: geti_sdk.rest_clients.dataset_client.DatasetClient
   :members:

.. autoclass:: geti_sdk.rest_clients.media_client.image_client.ImageClient
   :members:

.. autoclass:: geti_sdk.rest_clients.media_client.video_client.VideoClient
   :members:

.. autoclass:: geti_sdk.rest_clients.annotation_clients.annotation_client.AnnotationClient
   :members:

.. autoclass:: geti_sdk.rest_clients.configuration_client.ConfigurationClient
   :members:

.. autoclass:: geti_sdk.rest_clients.prediction_client.PredictionClient
   :members:

.. autoclass:: geti_sdk.rest_clients.model_client.ModelClient
   :members:

.. autoclass:: geti_sdk.rest_clients.training_client.TrainingClient
   :members:

.. autoclass:: geti_sdk.rest_clients.deployment_client.DeploymentClient
   :members:

"""

from .active_learning_client import ActiveLearningClient
from .annotation_clients import AnnotationClient
from .configuration_client import ConfigurationClient
from .dataset_client import DatasetClient
from .deployment_client import DeploymentClient
from .media_client import ImageClient, VideoClient
from .model_client import ModelClient
from .prediction_client import PredictionClient
from .project_client import ProjectClient
from .testing_client import TestingClient
from .training_client import TrainingClient

__all__ = [
    "AnnotationClient",
    "ConfigurationClient",
    "DatasetClient",
    "ProjectClient",
    "VideoClient",
    "ImageClient",
    "PredictionClient",
    "ModelClient",
    "TrainingClient",
    "DeploymentClient",
    "ActiveLearningClient",
    "TestingClient",
]



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from enum import Enum


class ConfigurationEntityType(Enum):
    """
    Enum representing the different configuration types on the Intel® Geti™ platform.
    """

    HYPER_PARAMETER_GROUP = "HYPER_PARAMETER_GROUP"
    COMPONENT_PARAMETERS = "COMPONENT_PARAMETERS"

    def __str__(self):
        """
        Return the string representation of the ConfigurationEntityType instance.
        """
        return self.value


class ParameterDataType(Enum):
    """
    Enum representing the different data types for configurable parameters on the
    Intel® Geti™ platform.
    """

    BOOLEAN = "boolean"
    FLOAT = "float"
    STRING = "string"
    INTEGER = "integer"

    def __str__(self):
        """
        Return the string representation of the ParameterDataType instance.
        """
        return self.value


class ParameterInputType(Enum):
    """
    Enum representing the different input types for configurable parameters on the
    Intel® Geti™ platform.
    """

    INPUT = "input"
    SELECTABLE = "selectable"

    def __str__(self):
        """
        Return the string representation of the ParameterInputType instance.
        """
        return self.value


class ConfigurableParameterType(Enum):
    """
    Enum representing the different types of configurable parameters on the
    Intel® Geti™ platform.
    """

    CONFIGURABLE_PARAMETERS = "CONFIGURABLE_PARAMETERS"
    PARAMETER_GROUP = "PARAMETER_GROUP"

    def __str__(self):
        """
        Return the string representation of the ConfigurableParameterType instance.
        """
        return self.value



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
Introduction
------------

These pages contain the documentation for the main SDK class,
:py:class:`~geti_sdk.sc_rest_client.Geti`.

The :py:class:`~geti_sdk.sc_rest_client.Geti` class implements convenience
methods for common operations that can be performed on the Intel® Geti™ cluster, such as
creating a project from a pre-existing dataset, downloading or uploading a project,
uploading an image and getting a prediction for it and creating a deployment for a
project.

For example, to download a project, simply do:

.. code-block:: python

   from geti_sdk import Geti

   geti = Geti(
     host="https://0.0.0.0", username="dummy_user", password="dummy_password"
   )

   geti.download_project(project_name="dummy_project")

For custom operations or more fine-grained control over the behavior, the
:py:mod:`~geti_sdk.rest_clients` subpackage should be used.

Module contents
---------------

.. autoclass:: geti_sdk.geti::Geti
   :no-members:

   .. rubric:: Project download and upload

   .. automethod:: download_project

   .. automethod:: upload_project

   .. automethod:: download_all_projects

   .. automethod:: upload_all_projects

   .. rubric:: Project creation from dataset

   .. automethod:: create_single_task_project_from_dataset

   .. automethod:: create_task_chain_project_from_dataset

   .. rubric:: Project deployment

   .. automethod:: deploy_project

   .. rubric:: Media upload and prediction

   .. automethod:: upload_and_predict_image

   .. automethod:: upload_and_predict_video

   .. automethod:: upload_and_predict_media_folder

"""

from .geti import Geti

__version__ = "2.0.1"

__all__ = ["Geti"]
